# 3rd-party
from lxml import etree
from lxml.builder import E 

# package modules
from .. import jxml as JXML
from .cfg import Resource, P_JUNOS_EXISTS, P_JUNOS_ACTIVE

class TemplateResource( Resource ):

  ##### -----------------------------------------------------------------------
  ##### PROPERTIES
  ##### -----------------------------------------------------------------------

  @property
  def active(self):
    """
      returns :True: if every named element in the XML is active
      returns :False: otherwise
    """
    _active = True

    for name in self._name:
      if not self.has[P_JUNOS_ACTIVE+'_'+name]:
        _active = False
        break

    return _active

  @property
  def exists(self):
    """
      returns :True: if every named element in the XML exists
      returns :False: otherwise
    """
    _exists = True

    for name in self._name:
      if not self.has[P_JUNOS_EXISTS+'_'+name]:
        _exists = False
        break

    return _exists
  
  ### -------------------------------------------------------------------------
  ### activate
  ### -------------------------------------------------------------------------

  def activate(self, **kvargs):
    """
      write config to activate resource; i.e. "activate ..."
    """
    #! removed check on active due to multi-key template
    #@@@ revisit

    names =  kvargs if len(kvargs) else self._name

    xml = self._xml_template_active( JXML.ACTIVATE, names )
    rsp = self._r_config_write_xml( xml )

    for name in names:
      self.has[P_JUNOS_ACTIVE+'_'+name] = True

    return True

  ### -------------------------------------------------------------------------
  ### deactivate
  ### -------------------------------------------------------------------------

  def deactivate(self, **kvargs):
    """
      write config to deactivate resource, i.e. "deactivate ..."
    """
    #! removed check on active due to multi-key template
    #@@@ revisit

    names =  kvargs if len(kvargs) else self._name

    xml = self._xml_template_active( JXML.DEACTIVATE, names )
    rsp = self._r_config_write_xml( xml )

    for name in names:
      self.has[P_JUNOS_ACTIVE+'_'+name] = False

    return True

  ### -------------------------------------------------------------------------
  ### delete
  ### -------------------------------------------------------------------------

  def delete(self, **kvargs):
    """
      remove configuration from Junos device
    """
    # ! removed check on exists due to multi-key template
    # ! @@ revisit

    names =  kvargs if len(kvargs) else self._name

    xml = self._xml_template_delete( names )
    rsp = self._r_config_write_xml( xml )

    for name in names:
      self.has[P_JUNOS_EXISTS+'_'+name] = False

    return True

  ### -------------------------------------------------------------------------
  ### rename
  ### -------------------------------------------------------------------------

  def rename(self, new_name=None, **kvargs):
    """
      Not supported (for now) - raise RuntimeError
    """
    raise RuntimeError("rename not supported on %s" % self.__class__.__name__)

  ### -------------------------------------------------------------------------
  ### reorder
  ### -------------------------------------------------------------------------

  def reorder( self, **kvargs ):
    """
      Not supported - raises RuntimeError
    """
    raise RuntimeError("reorder not supported on %s" % self.__class__.__name__)

  ##### -----------------------------------------------------------------------
  ##### ~private~ OVERLOADING
  ##### -----------------------------------------------------------------------

  def _r_config_read_xml(self):
    """
    ~! OVERLOADS :Resource: !~
      read the resource config from the Junos device
    """
    return self._junos.rpc.get_config(self._xml_template_read())

  def _xml_build_change(self):
    """
    ~! OVERLOADS :Resource !~
      run the templater to produce the XML for change.  this 
      method is invoked by the parent :Resource: during the 
      invocation of the :write():
    """
    return self._xml_template_write()

  ##### -----------------------------------------------------------------------
  ##### ~private~ resource helper methods
  ##### -----------------------------------------------------------------------

  def _r_has_init(self):
    """
      initializes the :has: attribute
    """
    self.has.clear()
    for name in self._name: 
      self.has[P_JUNOS_EXISTS+'_'+name] = False
      self.has[P_JUNOS_ACTIVE+'_'+name] = False

  def _r_set_active(self, my_props, value):
    """
      sets each of the :has: name P_JUNOS_ACTIVE
      properties to :value:
    """
    for name in self._name: 
      my_props[P_JUNOS_ACTIVE+'_'+name] = value

  def _r_set_exists(self, my_props, value):
    """
      sets each of the name P_JUNOS_ACTIVE
      properties to in the :my_props: to :value:
    """
    for name in self._name: 
      my_props[P_JUNOS_EXISTS+'_'+name] = value

  def _r_has_xml_status( self, xml_ele, to_py ):
    """
      set the P_JUNOS_EXISTS and P_JUNOS_ACTIVE property 
      states for each item in the resource name as extracted
      from the XML configuration retrieved from the Junos device
    """
    for name in self._xpath_names:
      if xml_ele[name] is not None:        
        to_py[P_JUNOS_EXISTS+'_'+name] = True 
        active = False if xml_ele[name].attrib.has_key('inactive') else True
        to_py[P_JUNOS_ACTIVE+'_'+name] = active
      else:
        to_py[P_JUNOS_EXISTS+'_'+name] = False
        to_py[P_JUNOS_ACTIVE+'_'+name] = False    

  def _r_template_write_vars(self):
    """
      This method is called when the XML config-write template
      is being formed.  by default the simplest case is to combine
      all of the variables together.  more complicated template 
      resources can overload this method to do what they want.
    """
    t_vars = dict(self._name)
    t_vars.update( self.has )
    t_vars.update( self.should )
    return t_vars

  ##### -----------------------------------------------------------------------
  ##### ~private~ methods used to create XML structures from templates
  ##### -----------------------------------------------------------------------

  def _xml_template_active(self, cmd, names ):
    """
      create an XML structure that will activate/deactivate
      the named elements in the resource configuration
    """

    xml = self._xml_template_names_only()
    for name,xpath in self._xpath_names.items():
      if name not in names: continue
      ele = xml.find(xpath)
      ele.attrib.update(cmd)
    return xml

  def _xml_template_delete(self, names):
    """
      creates an XML structure that will delete
      the named elements in the resource configuration
    """

    xml = self._xml_template_names_only()
    for name,xpath in self._xpath_names.items():
      if name not in names: continue      
      ele = xml.find(xpath)
      ele.attrib.update( JXML.DEL )
    return xml

  def _xml_template_read(self):
    """
      create an XML structure from template that
      will be used to read the resource configuration 
      from the Junos device
    """
    t = self._j2_ldr.get_template(self._j2_rd )    
    return etree.XML(t.render(self._name))    

  def _xml_template_names_only( self ):
    """
      create an XML structure from template that will
      be used to read only the resource name values 
      from the Junos device
    """
    t = self._j2_ldr.get_template(self._j2_rd)
    return etree.XML(t.render(self._name, NAMES_ONLY=True))      

  def _xml_template_write(self):
    t_vars = self._r_template_write_vars()    
    t = self._j2_ldr.get_template( self._j2_wr ) 
    return etree.XML(t.render( t_vars ))
  



# 3rd-party modules
from lxml.builder import E 

# module packages
from ... import Resource
from .... import jxml as JXML

class NatProxyArp( Resource ):
  """
  [edit security nat proxy-arp interface <if_name> address <ip_prefix>]

  Resource namevar: 
    tuple(if_name, ip_prefix)

  Description:
    This resource allows you to add/remove proxy-arp entries for NAT.  At 
    this time, there are no managed properties, so you can simply add or
    remove entries by the name tuple(if_name, ip_prefix)

    For example, to select an entry directly:

      entry = NatProxyArp(jdev, ('reth0.213','198.18.11.5'))

    Or using the bind mechanism:
    
      jdev.bind(parp=NatProxyArp)
      entry = jdev.parp[('reth0.213', '198.18.11.5')]

    To create it, you need to use the 'touch' option when invoking
    write() since there are no properites for proxy-arp entries

      if not entry.exists: 
        entry.write(touch=True)

    And to remove the same entry:

      entry.delete()
  """

  def _xml_at_top(self):
    return E.security(E.nat(
      E('proxy-arp',
        E.interface(E.name( self._name[0] ),
          E.address(E.name( self._name[1]))
        )
      )
    ))

  ##### -----------------------------------------------------------------------
  ##### OVERLOADS
  ##### -----------------------------------------------------------------------

  def rename(self, name):
    """ UNSUPPORTED """
    raise RuntimeError("Unsupported for Resource: %s" % self.__class__.__name__)

  ##### -----------------------------------------------------------------------
  ##### XML read
  ##### -----------------------------------------------------------------------

  def _xml_at_res(self, xml):
    return xml.find('.//proxy-arp/interface')

  def _xml_to_py(self, as_xml, to_py ):
    Resource._r_has_xml_status( as_xml, to_py )

  ##### -----------------------------------------------------------------------
  ##### Resource List, Catalog
  ##### -- only executed by 'manager' resources
  ##### -----------------------------------------------------------------------

  def _r_list(self):
    raise RuntimeError("@@@ NEED TO IMPLEMENT!")

  def _r_catalog(self):
    raise RuntimeError("@@@ NEED TO IMPLEMENT!")



# 3rd-party
from lxml import etree

# local
from .table import Table

class OpTable(Table):

  ### -------------------------------------------------------------------------
  ### PUBLIC METHODS
  ### -------------------------------------------------------------------------

  def get(self, *vargs, **kvargs):
    """ 
    Retrieve the XML table data from the Device instance and
    returns back the Table instance - for call-chaining purposes.  

    If the Table was created with a :path: rather than a Device,
    then this method will load the XML from that file.  In this
    case, the *vargs, and **kvargs are not used.

    ALIAS: __call__    

    :vargs:
      [0] is the table :arg_key: value.  This is used so that
      the caller can retrieve just one item from the table without
      having to know the Junos RPC argument.

    :kvargs:
      these are the name/value pairs relating to the specific Junos
      XML command attached to the table.  For example, if the RPC 
      is 'get-route-information', there are parameters such as
      'table' and 'destination'.  Any valid RPC argument can be
      passed to :kvargs: to further filter the results of the :get():
      operation.  neato!

    NOTES:
      If you need to create a 'stub' for unit-testing
      purposes, you want to create a subclass of your table and 
      overload this methods.
    """
    self._clearkeys()

    if self._path is not None:
      # for loading from local file-path
      self.xml = etree.parse(self._path).getroot()
      return self

    argkey = vargs[0] if len(vargs) else None

    rpc_args = {}                     # create empty <dict>
    rpc_args.update(self.GET_ARGS)    # copy default args
    rpc_args.update(kvargs)           # copy caller provided args

    if hasattr(self, 'GET_KEY') and argkey is not None:
      rpc_args.update({self.GET_KEY: argkey })

    # execute the Junos RPC to retrieve the table
    self.xml = getattr(self.RPC,self.GET_RPC)(**rpc_args)

    # returning self for call-chaining purposes, yo!
    return self



# debuggin
from lxml import etree

# 3rd-party modules
from lxml.builder import E 

# module packages
from ... import jxml as JXML
from .. import Resource
from .shared_ab_addr import SharedAddrBookAddr
from .shared_ab_set import SharedAddrBookSet

class SharedAddrBook( Resource ):
  """
  [edit security address-book <name>]

  Resource <name>
    The address book name, string

  Manages:
    addr - SharedAddrBookAddr resources
    set  - SharedAddrBookAddrSet resources
  """
  PROPERTIES = [
    'description',
    '$addrs',         # read-only addresss
    '$sets',          # read-only address-sets
    'zone_list'       # attached zone
  ]
  def __init__(self, junos, name=None, **kvargs ):
    if name is None:
      # resource-manager
      Resource.__init__( self, junos, name, **kvargs )
      return

    self.addr = SharedAddrBookAddr( junos, parent=self )
    self.set = SharedAddrBookSet( junos, parent=self )
    self._manages = ['addr','set']
    Resource.__init__( self, junos, name, **kvargs )

  def _xml_at_top(self):
    return E.security(
      E('address-book', E.name( self._name ))
    )

  ##### -----------------------------------------------------------------------
  ##### XML reading
  ##### -----------------------------------------------------------------------
  
  def _xml_hook_read_begin(self, xml):
    ab = xml.find('.//address-book')
    ab.append(E('description'))
    ab.append(E('address', JXML.NAMES_ONLY))
    ab.append(E('address-set', JXML.NAMES_ONLY))
    ab.append(E('attach'))
    return True

  def _xml_at_res(self, xml):
    return xml.find('.//address-book')

  def _xml_to_py(self, as_xml, to_py ):
    Resource._r_has_xml_status( as_xml, to_py )
    Resource.copyifexists( as_xml, 'description', to_py)
    to_py['$addrs'] = [name.text for name in as_xml.xpath('address/name')]
    to_py['$sets'] = [name.text for name in as_xml.xpath('address-set/name')]

  ##### -----------------------------------------------------------------------
  ##### XML writing
  ##### -----------------------------------------------------------------------    

  def _xml_change_zone_list(self, xml):
    x_attach = E('attach')
    self._xml_list_property_add_del_names( x_attach,
      prop_name='zone_list', element_name='zone')
    xml.append(x_attach)
    return True

  ##### -----------------------------------------------------------------------
  ##### Manager List, Catalog
  ##### -----------------------------------------------------------------------

  def _r_list(self):
    raise RuntimeError("Need to implement!")

  def _r_catalog(self):
    raise RuntimeError("Need to implement!")



"""
This file contains the RunstatLoader class that is used to dynamically
create Runstat Table and View objects from a <dict> of data.  The <dict> can
originate from any kind of source: YAML, JSON, program.  For examples of YAML
refer to the .yml files in this jnpr.junos.op directory.
"""
# stdlib
from copy import deepcopy

# locally
from .factory_cls import *
from .viewfields import *

__all__ = ['FactoryLoader']

# internally used shortcuts

_VIEW = FactoryView
_FIELDS = ViewFields
_GET = FactoryOpTable 
_TABLE = FactoryTable
_CFGTBL = FactoryCfgTable

class FactoryLoader(object):
  """
  Used to load a <dict> of data that contains Table and View definitions.

  The primary method is :load(): which will return a <dict> of item-name and
  item-class definitions.  

  If you want to import these definitions directly into your namespace, (like a module)
  you would do the following:

    loader = RunstatLoader()
    catalog = loader.load( <catalog_dict> )
    globals().update( catalog )

  If you did not want to do this, you can access the items as the catalog.  For 
  example, if your <catalog_dict> contained a Table called MyTable, then you could do
  something like:

    MyTable = catalog['MyTable']
    table = MyTable(dev)
    table.get()
    ...
  """  
  def __init__(self):
    self._catalog_dict = None       # YAML data

    self._item_optables = []    # list of the get/op-tables
    self._item_cfgtables = []   # list of get/cfg-tables
    self._item_views = []        # list of views to build
    self._item_tables = []       # list of tables to build

    self.catalog = {}           # catalog of built classes

  ##### -----------------------------------------------------------------------
  ##### Create a View class from YAML definition
  ##### -----------------------------------------------------------------------

  def _fieldfunc_True(self, value_rhs):
    return lambda x: x == value_rhs

  def _fieldfunc_False(self, value_rhs):
    return lambda x: x != value_rhs

  def _add_dictfield(self, fields, f_name, f_dict, kvargs ):
    """ add a field based on its associated dictionary """
    # at present if a field is a <dict> then there is **one
    # item** - { the xpath value : the option control }.  typically
    # the option would be a bultin class type like 'int'
    # however, as this framework expands in capability, this
    # will be enhaced, yo!

    xpath, opt = f_dict.items()[0]       # get first/only key,value
    if 'flag' == opt: opt = 'bool'       # flag is alias for bool

    # first check to see if the option is a built-in Python
    # type, most commonly would be 'int' for numbers, like counters

    astype = __builtins__.get(opt) or globals().get(opt)
    if astype is not None:
      kvargs['astype'] = astype
      fields.astype( f_name, xpath, **kvargs)
      return

    # next check to see if this is a "field-function"
    # operator in the form "func=value", like "True=enabled"

    if isinstance(opt,str) and opt.find('=') > 0:
      field_cmd,value_rhs = opt.split('=')
      fn_field = '_fieldfunc_' + field_cmd
      if not hasattr(self, fn_field): 
        raise ValueError("Unknown field-func: '%'" % field_cmd)
      kvargs['astype'] = getattr(self,fn_field)(value_rhs)
      fields.astype( f_name, xpath, **kvargs )
      return

    raise RuntimeError("Dont know what to do with field: '%s'" % f_name)

  # ---[ END: _add_dictfield ] ------------------------------------------------

  def _add_view_fields(self, view_dict, fields_name, fields):
    """ add a group of fields to the view """
    fields_dict = view_dict[fields_name]
    try:
      # see if this is a 'fields_<group>' collection, and if so
      # then we automatically setup using the group mechanism
      mark = fields_name.index('_')
      group = {'group':fields_name[mark+1:]}
    except:
      # otherwise, no group, just standard 'fields'
      group={}

    for f_name, f_data in fields_dict.items():  
      # each field could have its own unique set of properties
      # so create a kvargs <dict> each time.  but copy in the
      # groups <dict> (single item) generically.
      kvargs = {}
      kvargs.update(group)

      if isinstance(f_data,dict):
        self._add_dictfield( fields, f_name, f_data, kvargs )
        continue

      if f_data in self._catalog_dict:
        # f_data is the table name
        cls_tbl = self.catalog.get(f_data, self._build_table( f_data ))
        fields.table( f_name, cls_tbl )
        continue        

      # if we are here, then it means that the field is a string value
      xpath = f_name if f_data is True else f_data
      fields.str(f_name,xpath,**kvargs)

  ### -------------------------------------------------------------------------

  def _build_view(self, view_name):
    """ build a new View definition """
    if view_name in self.catalog: return self.catalog[view_name]

    view_dict = self._catalog_dict[view_name]
    kvargs = { 'view_name' : view_name }

    # if there are field groups, then get that now.
    if 'groups' in view_dict: kvargs['groups'] = view_dict['groups']

    # if this view extends another ...
    if 'extends' in view_dict:
      base_cls =  self.catalog.get(view_dict['extends'])
      # @@@ should check for base_cls is None!
      kvargs['extends'] = base_cls

    fields = _FIELDS()
    fg_list = [name for name in view_dict if name.startswith('fields')]
    for fg_name in fg_list: 
      self._add_view_fields( view_dict, fg_name, fields )

    cls = _VIEW( fields.end, **kvargs )
    self.catalog[view_name] = cls
    return cls

  ##### -----------------------------------------------------------------------
  ##### Create a Get-Table from YAML definition
  ##### -----------------------------------------------------------------------

  def _build_optable( self, table_name):
    """ build a new Get-Table definition """
    if table_name in self.catalog: return self.catalog[table_name]

    tbl_dict = self._catalog_dict[table_name]
    kvargs = deepcopy(tbl_dict)

    rpc = kvargs.pop('rpc')
    kvargs['table_name'] = table_name

    if 'view' in tbl_dict:
      view_name = tbl_dict['view']
      cls_view = self.catalog.get( view_name, self._build_view( view_name ))
      kvargs['view'] = cls_view

    cls = _GET(rpc, **kvargs)
    self.catalog[table_name] = cls
    return cls

  ##### -----------------------------------------------------------------------
  ##### Create a Table class from YAML definition
  ##### -----------------------------------------------------------------------

  def _build_table(self, table_name ):
    """ build a new Table definition """
    if table_name in self.catalog: return self.catalog[table_name]

    tbl_dict = self._catalog_dict[table_name]

    table_item = tbl_dict.pop('item')
    kvargs = deepcopy(tbl_dict)
    kvargs['table_name'] = table_name

    if 'view' in tbl_dict:
      view_name = tbl_dict['view']
      cls_view = self.catalog.get( view_name, self._build_view( view_name ))
      kvargs['view'] = cls_view

    cls = _TABLE(table_item, **kvargs)
    self.catalog[table_name] = cls
    return cls

  def _build_cfgtable( self, table_name):
    """ build a new Config-Table definition """
    if table_name in self.catalog: return self.catalog[table_name]
    tbl_dict = self._catalog_dict[table_name]

    if 'view' in tbl_dict:
      # transpose name to class
      view_name = tbl_dict['view']      
      tbl_dict['view'] = self.catalog.get( view_name, self._build_view( view_name ))

    cls = _CFGTBL( table_name, tbl_dict )
    self.catalog[table_name] = cls
    return cls

  ##### -----------------------------------------------------------------------
  ##### Primary builders ...
  ##### -----------------------------------------------------------------------

  def _sortitems(self):
    for k,v in self._catalog_dict.items():
      if 'rpc' in v:
        self._item_optables.append(k)
      elif 'get' in v:
        self._item_cfgtables.append(k)
      elif 'view' in v:
        self._item_tables.append(k)
      else:
        self._item_views.append(k)

  def load( self, catalog_dict, envrion={} ):

    # load the yaml data and extract the item names.  these names will
    # become the new class definitions

    self._catalog_dict = catalog_dict
    self._sortitems()

    map( self._build_optable, self._item_optables )
    map( self._build_cfgtable, self._item_cfgtables )
    map( self._build_table, self._item_tables )
    map( self._build_view, self._item_views )

    return self.catalog


"""
This file contains variables defining the filesystem paths to various
directories within the Junos PyEZ microframework
"""
from os.path import dirname as _dirname
from os.path import join as _join

MODULEDIR = _dirname(__file__)
CFGDIR = _join(MODULEPATH,'cfg')
OPDIR = _join(MODULEPATH,'op')
CFGRODIR = _join(MODULEPATH,'cfgro')



from .chassis import chassis
from .routing_engines import routing_engines
from .personality import personality
from .swver import software_version
from .ifd_style import ifd_style
from .switch_style import switch_style
from .session import *
from .srx_cluster import srx_cluster

FACT_LIST = [
  chassis,
  routing_engines,  
  personality,  
  srx_cluster,
  software_version,
  ifd_style,
  switch_style,
  session
]

__all__ = ['FACT_LIST']


# stdlib
from inspect import isclass
from time import time
from datetime import datetime
import os

# 3rd-party
from lxml import etree

_TSFMT = "%Y%m%d%H%M%S"

class Table(object):
  ITEM_XPATH = None
  ITEM_NAME_XPATH = 'name'
  VIEW = None

  def __init__(self, dev=None, xml=None, path=None):
    """
    :dev: Device instance
    :xml: lxml Element instance 
    :path: file path to XML, to be used rather than :dev:
    """
    self._dev = dev 
    self.xml = xml
    self.view = self.VIEW
    self._key_list = []
    self._path = path

  ##### -------------------------------------------------------------------------
  ##### PROPERTIES
  ##### -------------------------------------------------------------------------    

  @property 
  def D(self):
    """ the Device instance """
    return self._dev

  @property 
  def RPC(self):
    """ the Device.rpc instance """
    return self.D.rpc

  @property
  def view(self):
    """ returns the current view assigned to this table """
    return self._view

  @view.setter
  def view(self, cls):
    """ assigns a new view to the table """
    if cls is None:
      self._view = None
      return

    if not isclass(cls):
      raise ValueError("Must be given RunstatView class")

    self._view = cls

  @property
  def hostname(self):
    return self.D.hostname

  @property 
  def is_container(self):
    """ 
    True if this table does not have records, but is a container of fields
    False otherwise
    """
    return self.ITEM_XPATH is None    

  @property
  def key_list(self):
    """ the list of keys, as property for caching """
    return self._key_list
  
  ##### -------------------------------------------------------------------------
  ##### PRIVATE METHODS
  ##### -------------------------------------------------------------------------  

  def _assert_data(self):
    if self.xml is None: raise RuntimeError("Table is empty, use get()")    

  def _keys_composite(self, xpath, key_list):
    """ composite keys return a tuple of key-items """
#    _tkey = lambda this: tuple([this.findtext(k) for k in key_list ])
    _tkey = lambda this: tuple([this.xpath(k)[0].text for k in key_list ])
    return [_tkey(item) for item in self.xml.xpath(xpath)]

  def _keys_simple(self, xpath):
    return [x.text.strip() for x in self.xml.xpath(xpath)]

  def _keyspec(self):
    """ returns tuple (keyname-xpath, item-xpath) """    
    return (self.ITEM_NAME_XPATH, self.ITEM_XPATH)

  def _clearkeys(self):
    self._key_list = []

  ##### -------------------------------------------------------------------------
  ##### PUBLIC METHODS
  ##### -------------------------------------------------------------------------    

  ## --------------------------------------------------------------------------
  ## keys
  ## --------------------------------------------------------------------------

  def _keys(self):
    """ return a list of data item keys from the Table XML """

    self._assert_data()
    key_value, xpath = self._keyspec()

    if isinstance(key_value, str):
      return self._keys_simple( xpath+'/'+key_value)

    if not isinstance( key_value, list ): 
      raise RuntimeError("What to do with key, table:'%'" % self.__class__.__name__)

    # ok, so it's a list, which means we need to extract tuple values
    return self._keys_composite( xpath, key_value )

  def keys(self):
    # if the key_list has been cached, then use it
    if len(self.key_list): return self.key_list

    # otherwise, build the list of keys into the cache
    self._key_list = self._keys()
    return self._key_list

  ## --------------------------------------------------------------------------
  ## values
  ## --------------------------------------------------------------------------

  def values(self):
    """ returns list of table entry items() """

    self._assert_data()
    if self.view is None:
      # no View, so provide XML for each item
      return [this for this in self]
    else:
      # view object for each item
      return [this.items() for this in self]

  ## --------------------------------------------------------------------------
  ## items
  ## --------------------------------------------------------------------------

  def items(self):
    """ returns list of tuple(name,values) for each table entry """
    return zip(self.keys(), self.values())

  ## --------------------------------------------------------------------------
  ## get - loads the data from source
  ## --------------------------------------------------------------------------

  def get(self, *vargs, **kvargs):
    # implemented by either OpTable or CfgTable
    # @@@ perhaps this should raise an exception rather than just 'pass', ??
    pass

  ## --------------------------------------------------------------------------
  ## savexml - saves the table XML to a local file
  ## --------------------------------------------------------------------------

  def savexml(self, path, hostname=False, timestamp=False, append=None):
    """
    Save a copy of the table XML data to a local file.  The name of the
    output file (:path:) can include the name of the Device host, the
    timestamp of this action, as well as any user-defined appended value.
    These 'add-ons' will be added to the :path: value prior to the file
    extension in the order (hostname,timestamp,append), separated by
    underscore (_).

    For example, if both hostname=True and append='BAZ1', then when
    :path: = '/var/tmp/foo.xml' and the Device.hostname is "srx123", the 
    final file-path will be "/var/tmp/foo_srx123_BAZ1.xml"

    :path:
      file-path to write the XML file on the local filesystem

    :hostname:
      if True, will append the hostname to the :path:

    :timestamp:
      if True, will append the timestamp to the :path: using the default
        timestamp format
      if <str> the timestamp will use the value as the timestamp format as
        defied by strftime()

    :append:
      any <str> value that you'd like appended to the :path: value preceeding the
      filename extension.
    """
    fname, fext = os.path.splitext(path)

    if hostname is True: fname += "_%s" % self.D.hostname

    if timestamp is not False:
      tsfmt = _TSFMT if timestamp is True else timestamp
      tsfmt_val = datetime.fromtimestamp(time()).strftime(tsfmt)
      fname += "_%s" % tsfmt_val

    if append is not None: fname += "_%s" % append

    path = fname + fext
    return etree.ElementTree(self.xml).write(file(path,'w'))

  ##### -------------------------------------------------------------------------
  ##### OVERLOADS
  ##### -------------------------------------------------------------------------    

  __call__ = get

  def __repr__(self):
    cls_name = self.__class__.__name__
    source = self.D.hostname if self.D is not None else self._path

    if self.xml is None:
      return "%s:%s - Table empty" % (cls_name, source)
    else:
      n_items = len(self.keys())
      return "%s:%s: %s items" % (cls_name, source, n_items)

  def __len__(self):
    self._assert_data()    
    return len(self.keys())

  def __iter__(self):
    """ iterate over each time in the table """
    self._assert_data()

    as_xml = lambda table,view_xml: view_xml
    view_as = self.view or as_xml

    for this in self.xml.xpath(self.ITEM_XPATH):
      yield view_as( self, this )    

  def __getitem__(self, value):
    """
    returns a table item.  if a table view is set (should be by default) then
    the item will be converted to the view upon return.  if there is no table 
    view, then the XML object will be returned.

    :value:
      when it is a <string>, this will perform a select based on the key-name
      when it is a <tuple>, this will perform a select based on the compsite key-name
      when it is an <int>, this will perform a select based by position, like <list>
        [0] is the first item 
        [-1] is the last item
      when it is a <slice> then this will return a <list> of View widgets
    """
    self._assert_data()

    if isinstance(value,int):
      # if selection by index, then grab the key at this index and
      # recursively call this method using that key, yo!
      return self.__getitem__(self.key_list[value])

    if isinstance(value,slice):
      # implements the 'slice' mechanism
      return [self.__getitem__(key) for key in self.key_list[value]]

    # ---[ get_xpath ] --------------------------------------------------------

    def get_xpath(find_value):
      namekey_xpath, item_xpath = self._keyspec()      
      xnkv = '[normalize-space({})="{}"]'

      if isinstance(find_value,str):
        # find by name, simple key
        return item_xpath + xnkv.format( namekey_xpath, find_value)

      if isinstance(find_value,tuple):
        # composite key (value1, value2, ...) will create an
        # iterative xpath of the fmt statement for each key/value pair
        xpf = ''.join([xnkv.format(k.replace('_','-'),v) for k,v in zip(namekey_xpath, find_value)])
        return item_xpath + xpf    

    # ---[END: get_xpath ] --------------------------------------------------------

    found = self.xml.xpath(get_xpath( value ))
    if not len(found): return None

    as_xml = lambda table,view_xml: view_xml
    use_view = self.view or as_xml

    return use_view( table=self, view_xml=found[0] ) 

  def __contains__(self,key):
    """ membership for use with 'in' """
    return bool(key in self.key_list)


# 3rd-party
from lxml.builder import E

# local
from .. import Resource
from ... import JXML
from .base import PhyPortBase

class PhyPortSwitch( PhyPortBase ):

  PORT_SPEED = {
    'auto' : 'auto-negotiation',
    '10m' : 'ethernet-10m',
    '100m' : 'ethernet-100m',
    '1g': 'ethernet-1g'
  }

  ##### -----------------------------------------------------------------------
  ##### XML readers
  ##### -----------------------------------------------------------------------

  def _xml_to_py(self, has_xml, has_py ):
    PhyPortBase._xml_to_py(self, has_xml, has_py )

    # speed, duplex, loopback are all under 'ether-options'
    ethopts = has_xml.find('ether-options')
    if ethopts is None: return

    if ethopts.find('loopback') is not None: has_py['loopback'] = True

    speed = ethopts.find('speed')
    if speed is not None:
      # take the first child element
      has_py['speed'] = speed[0].tag
      PhyPortBase._set_invert(has_py,'speed', self.PORT_SPEED)      

    Resource.copyifexists(ethopts,'link-mode', has_py, 'duplex')
    if 'duplex' in has_py: PhyPortBase._set_invert(has_py, 'duplex', self.PORT_DUPLEX)

  ##### -----------------------------------------------------------------------
  ##### XML writers
  ##### -----------------------------------------------------------------------

  def _xml_hook_build_change_begin(self, xml):
    if any([ this in self.should for this in ['speed','duplex','loopback']]):
      self._ethopts = E('ether-options')
      xml.append(self._ethopts)

  def _xml_change_speed(self, xml):
    speed_tag = self.PORT_SPEED.get(self.speed)
    add_this = E.speed( JXML.DEL ) if speed_tag is None else E.speed(E( speed_tag ))
    self._ethopts.append( add_this )
    return True

  def _xml_change_duplex(self, xml):
    value = self.PORT_DUPLEX.get(self.duplex)
    Resource.xml_set_or_delete(self._ethopts, 'link-mode', value)
    return True

  def _xml_change_loopback(self, xml):
    self._ethopts.append(Resource.xmltag_set_or_del('loopback', self.loopback))
    return True


from lxml.builder import E
from .table import Table

class CfgTable(Table):

  ##### -----------------------------------------------------------------------
  ##### CONSTRUCTOR
  ##### -----------------------------------------------------------------------

  def __init__(self, dev):
    Table.__init__(self, dev)       # call parent constructor

    self._data_dict = self.DEFINE  # crutch
    self.ITEM_NAME_XPATH = self._data_dict.get('key','name')    
    self.ITEM_XPATH = self._data_dict['get']
    self.view = self._data_dict.get('view')

  ##### -----------------------------------------------------------------------
  ##### PROPERTIES
  ##### -----------------------------------------------------------------------

  @property 
  def required_keys(self):
    """ 
    return a list of the keys required when invoking :get(): 
    and :get_keys(): 
    """
    return self._data_dict.get('required_keys')

  @property 
  def keys_required(self):
    """ True/False - if this Table requires keys """
    return self.required_keys is not None

  ##### -----------------------------------------------------------------------
  ##### PRIVATE METHODS
  ##### -----------------------------------------------------------------------

  def _buildxml( self, namesonly=True ):
    """
    Return an lxml Element starting with <configuration> and comprised
    of the elements specified in the xpath expression.

    For example, and xpath = 'interfaces/interface' would produce:

     <configuration>
        <interfaces>
          <interface/>
        </interfaces>
      </configuration>

    If :namesonly: is True, then the XML will be encoded only to
    retrieve the XML name keys at the 'end' of the XPath expression.

    This return value can then be passed to dev.rpc.get_config() 
    to retrieve the specifc data
    """
    xpath = self._data_dict['get']    
    top = E('configuration')
    dot = top
    for name in xpath.split('/'):
      dot.append(E(name))
      dot = dot[0]

    if namesonly is True: dot.attrib['recurse'] = 'false'
    return top

  def _grindkey(self, key_xpath, key_value ):
    """ returns list of XML elements for key values """
    simple = lambda: [E(key_xpath.replace('_','-'), key_value)]
    composite = lambda: [E(xp.replace('_','-'),xv) for xp,xv in zip(key_xpath, key_value)]
    return simple() if isinstance(key_xpath,str) else composite()

  def _encode_requiredkeys(self, get_cmd, kvargs):
    """ 
    used to encode the required_keys values into the XML get-command.
    each of the required_key=<value> pairs are defined in :kvargs:
    """
    rqkeys = self._data_dict['required_keys']
    for key_name in self.required_keys:
      # create an XML element with the key/value 
      key_value = kvargs.get(key_name)
      if key_value is None: 
        raise ValueError("Missing required-key: '%s'" % key_name)
      key_xpath = rqkeys[key_name]
      add_keylist_xml = self._grindkey(key_xpath, key_value)

      # now link this item into the XML command, where key_name
      # designates the XML parent element
      dot = get_cmd.find('.//' + key_name.replace('_','-'))
      if dot is None:
        raise RuntimeError("Unable to find parent XML for key: '%'" % key_name )
      for _at,_add in enumerate(add_keylist_xml): dot.insert(_at,_add)

  def _encode_namekey(self, get_cmd, dot, namekey_value):
    """
    encodes the specific namekey_value into the get command so that the 
    returned XML configuration is the complete hierarchy of data. 
    """
    namekey_xpath = self._data_dict.get('key','name')
    keylist_xml = self._grindkey(namekey_xpath, namekey_value)
    for _add in keylist_xml: dot.append(_add)

  def _encode_getfields( self, get_cmd, dot ):
    for field_xpath in self._data_dict['get_fields']:
      dot.append(E(field_xpath))

  def _keyspec(self):
    """ returns tuple (keyname-xpath, item-xpath) """    
    return (self._data_dict.get('key','name'), self._data_dict['get'])
        
  ### -------------------------------------------------------------------------
  ### get - retrieve Table data
  ### -------------------------------------------------------------------------

  def get(self, *vargs, **kvargs):
    """
    Retrieve configuration data for this table.  By default only the 
    keys of the table is loaded.  This behavior can be overriden by 
    either requesting a sepcifc table item, or by calling with
    kvargs['values']=True

    :vargs:
      [0] identifies a unique item in the table, 
      same as calling with :kvargs['key']: value

    RESERVED :kvargs:
      'values' - True/False*, when set to True will cause all data item
      values to be retrieved, not just the name-keys.  

      'key' - used to retrieve the contents of the table record, and not 
      just the list of keys (default) [when namesonly=True]
    """
    if self.keys_required is True and not len(kvargs):
      raise ValueError("This table has required-keys\n", self.required_keys)

    self._clearkeys()
    
    # determine if we need to get only the names of keys, or all of the
    # hierarchical data from the config.  The caller can explicity set
    # :namesonly: in the call, or request a specific items.  A specific
    # item can be identified either by vargs[0] or kvargs['key'].
    #
    # a caller might want to *botH* specify a named item, and namesonly=True
    # simply to test for the existance of an item in the config, yo!
    # this 'featuure' doesn't work yet, but looking into it. 

    b_values = kvargs.get('values')
    namesonly = (not (len(vargs) or ('key' in kvargs))) if b_values is None else not(b_values)

    get_cmd = self._buildxml(namesonly=namesonly)    

    # if this table requires additional keys, for the hierarchical use-cases
    # then make sure these are provided by the caller.  then encode them
    # into the 'get-cmd' XML

    if self.keys_required is True: 
      self._encode_requiredkeys( get_cmd, kvargs )

    try:
      # see if the caller provided a named item.  this must
      # be an actual name of a thing, and not an index number.
      # ... at least for now ...
      named_item = kvargs.get('key') or vargs[0]
      dot = get_cmd.find( self._data_dict['get'] )      
      self._encode_namekey( get_cmd, dot, named_item )

      if 'get_fields' in self._data_dict:
        self._encode_getfields( get_cmd, dot )

    except:
      # caller not requesting a sepcific table item
      pass

    self._get_cmd = get_cmd   # for debug purposes

    # retrieve the XML configuration
    self.xml = self.RPC.get_config( get_cmd )

    # return self for call-chaining, yo!
    return self



  ##### -----------------------------------------------------------------------
  ##### OVERLOADS
  ##### -----------------------------------------------------------------------


# 3rd-party modules
from lxml.builder import E 

# module packages
from ... import jxml as JXML
from .. import Resource

class ZoneAddrBookSet( Resource ):
  """
  [edit security zone security-zone <zone> address-book address-set <name>]

  Resource name: str
    <name> is the name of the address sets

  Managed by: ZoneAddrBook
    <zone> is the name of the secuirty zone
  """

  PROPERTIES = [
    'description',
    'addr_list',        # list of address items
    'set_list',         # sets can contain a list of sets
  ]

  def _xml_at_top(self):
    return E.security(E.zones(
      E('security-zone', 
        E.name(self.P._name),
        E('address-book', 
          E('address-set',E.name(self._name))
        )
      )
    ))

  ##### -----------------------------------------------------------------------
  ##### XML reading
  ##### -----------------------------------------------------------------------

  def _xml_at_res(self, xml):
    return xml.find('.//address-book/address-set')

  def _xml_to_py(self, as_xml, to_py ):
    Resource._r_has_xml_status( as_xml, to_py )
    Resource.copyifexists( as_xml, 'description', to_py )
    to_py['addr_list'] = [name.text for name in as_xml.xpath('address/name')]
    to_py['set_list'] = [name.text for name in as_xml.xpath('address-set/name')]

  ##### -----------------------------------------------------------------------
  ##### XML writing
  ##### -----------------------------------------------------------------------

  def _xml_change_addr_list( self, xml ):
    self._xml_list_property_add_del_names( xml,
      prop_name='addr_list', element_name='address')
    return True

  def _xml_change_set_list( self, xml ):
    self._xml_list_property_add_del_names( xml,
      prop_name='set_list',element_name='address-set')
    return True

  ##### -----------------------------------------------------------------------
  ##### Manager List, Catalog
  ##### -----------------------------------------------------------------------

  def _r_list(self):
    # list of address-book address-sets.  this list is managed by the
    # parent object, so just use that, yo!
    self._rlist = self.P['$sets']

  def _r_catalog(self):
    get = E.security(E.zones(
      E('security-zone', 
        E.name(self.P._name),
        E('address-book', 
          E('address-set')
        )
      )
    ))
    got = self.D.rpc.get_config(get)
    for adrset in got.xpath('.//address-set'):
      name = adrset.find('name').text
      self._rcatalog[name] = {}
      self._xml_to_py( adrset, self._rcatalog[name] )




# 3rd-party modules
from lxml.builder import E 

# module packages
from .. import Resource
from ... import jxml as JXML

class Application( Resource ):
  """
  [edit applications application <name>]

  Resource name: str
    <name> is the application item name
  """

  PROPERTIES = [
    'description',
    'protocol',
    'dest_port',
    'timeout'
  ]

  ##### -----------------------------------------------------------------------
  ##### XML readers
  ##### -----------------------------------------------------------------------

  def _xml_at_top(self):
    return E.applications(E.application(E.name( self._name )))

  def _xml_at_res(self, xml):
    return xml.find('.//application')

  def _xml_to_py(self, has_xml, has_py ):
    Resource._r_has_xml_status( has_xml, has_py )
    Resource.copyifexists( has_xml, 'description', has_py )
    has_py['protocol'] = has_xml.findtext('protocol')
    Resource.copyifexists( has_xml, 'destination-port', has_py, 'dest_port')
    Resource.copyifexists( has_xml, 'inactivity-timeout', has_py, 'timeout')
    if 'timeout' in has_py and has_py['timeout'] != 'never':
      has_py['timeout'] = int(has_py['timeout'])

  ##### -----------------------------------------------------------------------
  ##### XML property writers
  ##### -----------------------------------------------------------------------

  def _xml_change_protocol( self, xml ):
    xml.append(E.protocol(self['protocol']))
    return True

  def _xml_change_dest_port( self, xml ):
    """
    destination-port could be a single value or a range.
    handle the case where the value could be provided as either
    a single int value or a string range, e.g. "1-27"
    """
    value = self['dest_port']
    if isinstance(value,int): value = str(value)
    xml.append(E('destination-port', value))
    return True

  def _xml_change_timeout( self, xml ):
    xml.append(E('inactivity-timeout', str(self['timeout'])))
    return True

  ##### -----------------------------------------------------------------------
  ##### Manager List, Catalog
  ##### -----------------------------------------------------------------------

  def _r_list(self):
    get = E.applications(E.application( JXML.NAMES_ONLY ))
    got = self.R.get_config( get )
    self._rlist = [app.text for app in got.xpath('applications/application/name')]

  def _r_catalog(self):
    get = E.applications(E.application())
    got = self.R.get_config( get )
    for app in got.xpath('applications/application'):
      name = app.findtext('name')
      self._rcatalog[name] = {}
      self._xml_to_py( app, self._rcatalog[name] )


from lxml import etree
from . import jxml

class RpcError( Exception ):

  def __init__( self, cmd=None, rsp=None, errs=None ):
    """
      :cmd: is the rpc command
      :rsp: is the rpc response (after <rpc-reply>)
      :errs: is a list of <rpc-error> elements
    """
    self.cmd = cmd
    self.rsp = rsp
    self.errs = errs

  def __repr__(self):
    """
      pprints the response XML attribute
    """
    if None != self.rsp:
      return etree.tostring(self.rsp, pretty_print=True)

class CommitError( RpcError ):
  def __init__(self,cmd = None, rsp=None, errs=None):
    RpcError.__init__( self, cmd, rsp, errs )
    self.rpc_error = jxml.rpc_error( rsp )

class LockError( RpcError ):
  def __init__(self, rsp):
    RpcError.__init__(self, rsp=rsp)
    self.rpc_error = jxml.rpc_error( rsp )

class UnlockError( RpcError ):
  def __init__(self, rsp):
    RpcError.__init__(self, rsp=rsp)
    self.rpc_error = jxml.rpc_error( rsp )



from . import RunstatMaker as RSM

##### -------------------------------------------------------------------------
##### nice illustration of using the same View on different Tables
##### -------------------------------------------------------------------------

AlarmTableView = RSM.View(RSM.Fields()
  .int('count', 'alarm-summary/active-alarm-count')
  .table('list', RSM.Table('alarm-detail',
    key='alarm-short-description',
    view=RSM.View(RSM.Fields()
      .str('time','alarm-time')
      .int('time_epoc', 'alarm-time/@seconds')
      .str('severity''alarm-class')
      .str('type''alarm-type')
      .str('description','alarm-description')
      .str('brief','alarm-short-description')
      .end))
  )
  .end
)

###> show system alarms

SysAlarmTable = RSM.GetTable('get-system-alarm-information',
  key = None, view = AlarmTableView )

###> show chassis alarams

ChassisAlarmTable = RSM.GetTable('get-alarm-information',
  key = None, view = AlarmTableView )


# 3rd-party modules
from lxml.builder import E 

# module packages
from .. import Resource
from .policy_rule import PolicyRule
from ... import jxml as JXML

class PolicyContext( Resource ):
  """
  [edit security policy from-zone <from_zone> to-zone <to_zone>]

  Resource name: tuple(from_zone, to_zone)
    <from_zone>  is the name of the From zone
    <to_zone> is the name of the To zone

  Manages Resources:
    rule, PolicyRule 
  """

  PROPERTIES = [
    '$rules',
    '$rules_count'
  ]

  def __init__(self, junos, name=None, **kvargs ):
    if name is None:
      # resource-manager
      Resource.__init__( self, junos, name, **kvargs )
      return

    # specific instance will manage policy rules    
    self.rule = PolicyRule( junos, M=self, parent=self )
    self._manages = ['rule']    
    self._name_from_zone = name[0]
    self._name_to_zone = name[1]
    Resource.__init__(self,junos,name,**kvargs)

  def _xml_at_top(self):
    return E.security( E.policies(
      E.policy(
        E('from-zone-name', self._name_from_zone),
        E('to-zone-name', self._name_to_zone)
      )))

  ### -------------------------------------------------------------------------
  ### XML reading
  ### -------------------------------------------------------------------------

  def _xml_config_read(self):
    """
      ~! OVERLOADS !~
    """
    xml = self._xml_at_top()
    xml.find('.//policy').append(E.policy(JXML.NAMES_ONLY))
    return self._junos.rpc.get_config( xml )

  def _xml_at_res(self, xml):
    return xml.find('.//policy')

  def _xml_to_py(self, as_xml, to_py ):
    Resource._r_has_xml_status( as_xml, to_py )
    to_py['$rules'] = [policy.text for policy in as_xml.xpath('.//policy/name')]
    to_py['$rules_count'] = len(to_py['$rules'])

  ##### -----------------------------------------------------------------------
  ##### Resource List, Catalog
  ##### -- only executed by 'manager' resources
  ##### -----------------------------------------------------------------------

  def _r_list(self):
    """
      build the policy context resource list from the command:
      > show security policies zone-context
    """
    got = self._junos.rpc.get_firewall_policies( zone_context=True )

    for pc in got.xpath('//policy-zone-context/policy-zone-context-entry'):
      from_zone = pc.find('policy-zone-context-from-zone').text
      to_zone = pc.find('policy-zone-context-to-zone').text      
      self._rlist.append(( from_zone, to_zone ))

  def _r_catalog(self):
    got = self._junos.rpc.get_firewall_policies( zone_context=True )

    for pc in got.xpath('//policy-zone-context/policy-zone-context-entry'):
      from_zone = pc.find('policy-zone-context-from-zone').text
      to_zone = pc.find('policy-zone-context-to-zone').text
      count = int(pc.find('policy-zone-context-policy-count').text)
      self._rcatalog[(from_zone,to_zone)] = {'$rules_count': count }



"""
facts['srx_cluster'] = True | False 
  if personality.startswith('SRX')
  device responds to the "show" command to extract the data

Facts Required:
  personality
"""

def srx_cluster(junos,facts):
  if not facts['personality'].startswith('SRX'): return

  # we should check the 'cluster status' on redundancy group 0 to see who is 
  # master.  we use a try/except block for cases when SRX is not clustered

  try:
    cluster_st = junos.rpc.get_chassis_cluster_status(redundancy_group="0")     
    primary = cluster_st.xpath('.//redundancy-group-status[.="primary"]')[0]
    node = primary.xpath('preceding-sibling::device-name[1]')[0].text.replace('node','RE')
    if node not in facts['master']: facts['master'].append(node)
    facts['srx_cluster'] = True
  except:
    facts['srx_cluster'] = False
    pass


# debuggin
from lxml import etree

# 3rd-party modules
from lxml.builder import E 

# module packages
from ... import jxml as JXML
from .. import Resource
from .zone_ab import ZoneAddrBook
from .zone_ifs import ZoneInterface, HostInbSvcMixin

class Zone( HostInbSvcMixin, Resource ):
  """
  [edit security zone security-zone <zone>]

  Resource Name: str
    The zone name 

  Manages resources:
    ifs, ZoneInterface
    ab, ZoneAddressBook
  """

  PROPERTIES = [
    'description',
    'services',
    'protocols',
    '$ifs_list'
  ]

  def __init__(self, junos, name=None, **kvargs ):
    if name is None:
      # resource-manager
      Resource.__init__( self, junos, name, **kvargs )
      return

    self.ifs = ZoneInterface( junos, parent=self )
    self.ab = ZoneAddrBook( junos, name, parent=self )
    self._manages = ['ifs','ab']
    Resource.__init__( self, junos, name, **kvargs )

  ##### -----------------------------------------------------------------------
  ##### XML reading
  ##### -----------------------------------------------------------------------

  def _xml_at_top(self):
    return E.security(E.zones(
      E('security-zone',E.name( self._name ))
    ))

  def _xml_hook_read_begin(self, xml):
    e = xml.find('.//security-zone')
    e.append(E('host-inbound-traffic'))
    e.append(E('description'))
    e.append(E('interfaces'))
    return True

  def _xml_at_res(self, xml):
    return xml.find('.//security-zone')

  def _xml_to_py(self, as_xml, to_py ):
    Resource._r_has_xml_status( as_xml, to_py )
    Resource.copyifexists( as_xml, 'description', to_py)
    e = as_xml.xpath('host-inbound-traffic/system-services/name')
    to_py['services'] = [n.text for n in e]
    e = as_xml.xpath('host-inbound-traffic/protocols/name')
    to_py['protocols'] = [n.text for n in e]
    to_py['$ifs_list'] = [name.text for name in as_xml.xpath('interfaces/name')]
  
  ##### -----------------------------------------------------------------------
  ##### XML writing
  ##### -----------------------------------------------------------------------

  # handled by mixin, yo!

  ##### -----------------------------------------------------------------------
  ##### Manager List, Catalog
  ##### -----------------------------------------------------------------------

  def _r_list(self):
    got = self.R.get_zones_information(terse=True)
    zones = got.findall('zones-security/zones-security-zonename')
    self._rlist = [zone.text for zone in zones]
    self._rlist.remove('junos-host')

  # using Resource._r_catalog()


from lxml.builder import E

from .util import Util
from .start_shell import StartShell

class FS(Util):
  """
  Filesystem (FS) utilities:

    cat - show the contents of a file
    checksum - calculate file checksum (md5,sha256,sha1)
    copy - local file copy (not scp)
    cwd - change working directory
    ls - return file/dir listing
    mkdir - create a directory
    pwd - get working directory
    rename - local file rename
    rm - local file delete
    rmdir - remove a directory
    stat - return file/dir information
    storage_usage - return storage usage
    storage_cleanup - perform storage storage_cleanup
    storage_cleanup_check - returns a list of files to remove at cleanup
    symlink - create a symlink
    tgz - tar+gzip a directory

  NOTES:
    The following methods require 'start shell' priveldges:
    [mkdir, rmdir, symlink]
  """

  ### -------------------------------------------------------------------------
  ### cat - show file contents
  ### -------------------------------------------------------------------------

  def cat(self,path):
    """
    returns the contents of the file :path:
    """
    try:
      rsp = self._dev.rpc.file_show(filename=path)
    except:
      return None
    return rsp.text

  ### -------------------------------------------------------------------------
  ### cwd - change working directory
  ### -------------------------------------------------------------------------

  def cwd(self, path):
    """
    change working directory to path
    """
    self._dev.rpc.set_cli_working_directory(directory=path)

  ### -------------------------------------------------------------------------
  ### pwd - return current working directory
  ### -------------------------------------------------------------------------

  def pwd(self):
    """
    returns the current working directory
    """
    rsp = self._dev.rpc(E.command("show cli directory"))
    return rsp.findtext('./working-directory')

  ### -------------------------------------------------------------------------
  ### checksum - compute file checksum
  ### -------------------------------------------------------------------------

  def checksum(self, path, calc='md5' ):
    """
    performs the checksum command on the given file path using the 
    required calculation method ['md5', 'sha256', 'sha1'] and returns
    the string value.  if the :path: is not found on the device, then
    None is returned.
    """
    cmd_map = {
      'md5' : self._dev.rpc.get_checksum_information,
      'sha256' : self._dev.rpc.get_sha256_checksum_information,
      'sha1' : self._dev.rpc.get_sha1_checksum_information
    }
    rpc = cmd_map.get(calc)
    if rpc is None: raise ValueError("Unknown calculation method: '%s'" % calc)
    try:
      rsp = rpc(path=path)
      return rsp.findtext('.//checksum').strip()
    except:
      # the only exception is that the path is not found
      return None

  @classmethod
  def _decode_file(cls, fileinfo):
    results = {}

    not_file = fileinfo.xpath('file-directory | file-symlink-target')
    if len(not_file):
      results['type'] = {'file-directory':'dir', 'file-symlink-target':'link'}[not_file[0].tag]
      if 'link' == results['type']:
        results['link'] = not_file[0].text.strip()
    else:    
      results['type'] = 'file'

    results['path'] = fileinfo.findtext('file-name').strip()
    results['owner'] = fileinfo.findtext('file-owner').strip()
    results['size'] = int(fileinfo.findtext('file-size'))
    fper = fileinfo.find('file-permissions')
    results['permissions'] = int(fper.text.strip())
    results['permissions_text'] = fper.get('format')
    fdate = fileinfo.find('file-date')
    results['ts_date'] = fdate.get('format')
    results['ts_epoc'] = fdate.text.strip()
    return results

  @classmethod
  def _decode_dir(cls, dirinfo, files=None):
    results = {}
    results['type'] = 'dir'
    results['path'] = dirinfo.get('name')
    if files is None: files = dirinfo.xpath('file-information')
    results['file_count'] = len(files)
    results['size'] = sum([int(f.findtext('file-size')) for f in files])
    return results

  ### -------------------------------------------------------------------------
  ### stat - file information
  ### -------------------------------------------------------------------------

  def stat(self,path):
    """
    Returns a dictionary of status information on the path, or None
    if the path does not exist.  

    @@@ MORE NEEDED @@@
    """
    rsp = self._dev.rpc.file_list(detail=True, path=path)

    # if there is an output tag, then it means that the path
    # was not found
    if rsp.find('output') is not None: return None

    # ok, so we've either got a directory or a file at 
    # this point, so decode accordingly

    xdir = rsp.find('directory')
    if xdir.get('name'): # then this is a directory path
      return FS._decode_dir(xdir)
    else:
      return FS._decode_file(xdir.find('file-information'))

  ### -------------------------------------------------------------------------
  ### ls - file/dir listing
  ### -------------------------------------------------------------------------

  def ls(self, path='.', brief=False, followlink=True):
    """
    File listing, returns a dict of file information.  If the
    path is a symlink, then by default (:followlink):) will
    recursively call this method to obtain the symlink specific
    information.
    """
    rsp = self._dev.rpc.file_list(detail=True, path=path)

    # if there is an output tag, then it means that the path
    # was not found, and we return :None:

    if rsp.find('output') is not None: return None

    xdir = rsp.find('directory')

    # check to see if the directory element has a :name:
    # attribute, and if it does not, then this is a file, and
    # decode accordingly.  If the file is a symlink, then we
    # want to follow the symlink to get what we want.

    if not xdir.get('name'): 
      results = FS._decode_file(xdir.find('file-information'))
      link_path = results.get('link')
      if not link_path: # then we are done
        return results
      else:
        return results if followlink is False else self.ls(path=link_path)

    # if we are here, then it's a directory, include information on all files
    files = xdir.xpath('file-information')
    results = FS._decode_dir(xdir, files)

    if brief is True:
      results['files'] = [ f.findtext('file-name').strip() for f in files ]
    else:
      results['files'] = {
        f.findtext('file-name').strip():FS._decode_file(f)
        for f in files
      }

    return results

  ### -------------------------------------------------------------------------
  ### storage_usage - filesystem storage usage
  ### -------------------------------------------------------------------------

  def storage_usage(self):
    rsp = self._dev.rpc.get_system_storage()

    _name = lambda fs: fs.findtext('filesystem-name').strip()

    def _decode(fs):
      r = {}
      r['mount'] = fs.find('mounted-on').text.strip()
      tb = fs.find('total-blocks')
      r['total'] = tb.get('format')        
      r['total_blocks'] = int(tb.text)
      ub = fs.find('used-blocks')
      r['used'] = ub.get('format')        
      r['used_blocks'] = int(ub.text)
      r['used_pct'] = fs.find('used-percent').text.strip()
      ab = fs.find('available-blocks')
      r['avail'] = ab.get('format')        
      r['avail_block'] = int(ab.text)
      return r

    return { _name(fs):_decode(fs) for fs in rsp.xpath('filesystem') }

  ### -------------------------------------------------------------------------
  ### storage_cleanup_check, storage_cleanip
  ### -------------------------------------------------------------------------

  @classmethod
  def _decode_storage_cleanup(cls,files):
    _name = lambda f: f.findtext('file-name').strip()
    def _decode(f):
      return {
        'size' : int(f.findtext('size')),
        'ts_date' : f.findtext('date').strip()
      }

    # return a dict of name/decode pairs for each file
    return { _name(f): _decode(f) for f in files }

  def storage_cleanup_check(self):
    """
    Perform the 'request system storage cleanup dry-run' command
    to return a :dict: of files/info that would be removed if 
    the cleanup command was executed.
    """
    rsp = self._dev.rpc.request_system_storage_cleanup(dry_run=True)
    files = rsp.xpath('file-list/file')
    return FS._decode_storage_cleanup(files)

  def storage_cleanup(self):
    """
    Perform the 'request system storage cleanup' command to remove
    files from the filesystem.  Return a :dict: of file name/info 
    on the files that were removed.
    """
    rsp = self._dev.rpc.request_system_storage_cleanup()
    files = rsp.xpath('file-list/file')
    return FS._decode_storage_cleanup(files)

  ### -------------------------------------------------------------------------
  ### rm - local file delete
  ### -------------------------------------------------------------------------
    
  def rm(self,path):
    """
    Performs a local file delete action, per Junos CLI command "file delete".
    If the file does not exist, then this return returns False.
    """
    # the return value from this RPC will return either True if the delete
    # was successful, or an XML structure otherwise.  So we can do a simple
    # test to provide the return result to the caller.
    rsp = self._dev.rpc.file_delete(path=path)
    return rsp == True

  ### -------------------------------------------------------------------------
  ### cp - local file copy
  ### -------------------------------------------------------------------------

  def cp(self, from_path, to_path):
    """
    Perform a local file copy where :from_path: and :to_path: can be any 
    valid Junos path argument.  Refer to the Junos "file copy" command
    documentation for details.

    Returns True if OK, False if file does not exist.
    """
    # this RPC returns True if it is OK.  If the file does not exist
    # this RPC will generate an RpcError exception, so just return False
    try:
      self._dev.rpc.file_copy(source=from_path, destination=to_path)
    except:
      return False
    return True
    
  ### -------------------------------------------------------------------------
  ### mv - local file rename
  ### -------------------------------------------------------------------------

  def mv(self, from_path, to_path):
    """
    Perform a local file rename function, same as "file rename" Junos CLI.
    """
    rsp = self._dev.rpc.file_rename(source=from_path,destination=to_path)
    return rsp == True

  def tgz(self, from_path, tgz_path):
    """
    create a file called :tgz_path: that is the tar-gzip of the given
    directory specified :from_path:
    """
    rsp = self._dev.rpc.file_archive(compress=True, 
      source=from_path, destination=tgz_path)

    # if the rsp is True, then the command executed OK.    
    if rsp is True: return True

    # otherwise, return the error string to the caller
    return rsp.text

  ### -------------------------------------------------------------------------
  ### !!!!! methods that use SSH shell commands, requires that the user 
  ### !!!!! has 'start shell' priveldges
  ### -------------------------------------------------------------------------

  def _ssh_exec(self,command):
    with StartShell(self._dev) as sh:
      got = sh.run(command)
      ok = sh.last_ok
    return (ok,got)

  def rmdir(self, path):
    """
    ~| REQUIRES SHELL PRIVELDGES |~
    executes the 'rmdir' command on path
    returns True if OK, or error string
    """
    results = self._ssh_exec( "rmdir %s" % path )
    return True if results[0] is True else ''.join(results[1][2:-1])

  def mkdir(self, path):  
    """
    ~| REQUIRES SHELL PRIVELDGES |~
    executes the 'mkdir -p' command on path
    returns True if OK, or error string
    """
    results = self._ssh_exec("mkdir -p %s" % path)
    return True if results[0] is True else ''.join(results[1][2:-1])

  def symlink(self, from_path, to_path):
    """
    ~| REQUIRES SHELL PRIVELDGES |~
    executes the 'ln -sf <from_path> <to_path>' command
    returns True if OK, or error string
    """
    results = self._ssh_exec("ln -sf %s %s" % (from_path, to_path))
    return True if results[0] is True else ''.join(results[1][2:-1])



# 3rd-party
from lxml.builder import E 

# module packages
from .. import Resource
from ... import jxml as JXML

class ApplicationSet( Resource ):
  """
  [edit applications application-set <name>]

  Resource name: str
    <name> is the application-set name
  """

  PROPERTIES = [
    'description',
    'app_list',     'app_list_adds',    'app_list_dels',
    'appset_list',  'appset_list_adds', 'appset_list_dels'
  ]

  def _xml_at_top(self):
    return E.applications(E('application-set', (E.name( self._name ))))

  def _xml_at_res(self, xml):
    return xml.find('.//application-set')

  def _xml_to_py(self, has_xml, has_py ):
    Resource._r_has_xml_status( has_xml, has_py )    
    Resource.copyifexists( has_xml, 'description', has_py )


    # each of the <application> elements
    app_list = [this.findtext('name') for this in has_xml.xpath('application')]
    set_list = [this.findtext('name') for this in has_xml.xpath('application-set')]

    if len(app_list): has_py['app_list'] = app_list
    if len(set_list): has_py['appset_list'] = set_list

  ##### -----------------------------------------------------------------------
  ##### XML property writers
  ##### -----------------------------------------------------------------------

  ### -------------------------------------------------------------------------
  ### application list
  ### -------------------------------------------------------------------------

  def _xml_change_app_list( self, xml ):
    self._xml_list_property_add_del_names( xml,
      prop_name='app_list',element_name='application')
    return True

  def _xml_change_app_list_adds( self, xml ):
    for this in self.should['app_list_adds']:
      xml.append(E.application(E.name(this)))
    return True

  def _xml_change_app_list_dels( self, xml ):
    for this in self.should['app_list_dels']:
      xml.append(E.application(JXML.DEL, E.name(this)))
    return True

  ### -------------------------------------------------------------------------
  ### application-set list
  ### -------------------------------------------------------------------------

  def _xml_change_appset_list( self, xml ):
    if None == self.should.get('appset_list'): self['appset_list'] = []

    (adds,dels) = Resource.diff_list( self.has.get('appset_list',[]), self.should['appset_list'])

    for this in adds: xml.append(E('application-set',E.name(this)))
    for this in dels: xml.append(E('application-set', JXML.DEL, E.name(this)))
    return True

  def _xml_change_appset_list_adds( self, xml ):
    for this in self.should['appset_list_adds']:
      xml.append(E('application-set',E.name(this)))
    return True    

  def _xml_change_appset_list_dels( self, xml ):
    for this in self.should['appset_list_dels']:
      xml.append(E('application-set', JXML.DEL, E.name(this)))
    return True    

  ##### -----------------------------------------------------------------------
  ##### Resource List, Catalog
  ##### -- only executed by 'manager' resources
  ##### -----------------------------------------------------------------------

  def _r_list(self):
    got = self.R.get_config(E.applications(E('application-set', JXML.NAMES_ONLY)))
    self._rlist = [ this.text for this in got.xpath('.//name')]

  def _r_catalog(self):
    got = self.R.get_config(E.applications(E('application-set')))
    for this in got.xpath('.//application-set'):
      name = this.findtext('name')
      self._rcatalog[name] = {}
      self._xml_to_py( this, self._rcatalog[name] )



import re

def switch_style( junos, facts ):
  persona = facts['personality']

  if persona in ['MX', 'SRX_HIGHEND']:
    style = 'BRIDGE_DOMAIN'
  elif persona in ['SWITCH', 'SRX_BRANCH']:
    model = facts['model']
    if re.match('firefly',model,re.IGNORECASE):
      style = 'NONE'
    elif re.match('^(EX9)|(EX43)', model):
      style = 'VLAN_L2NG'
    else:
      style = 'VLAN'
  else:
    style = 'NONE'

  facts['switch_style'] = style


from .app import Application
from .app_set import ApplicationSet
from .policy import PolicyContext
from .policy_rule import PolicyRule
from .zone import Zone
from .shared_ab import SharedAddrBook
from .zone_ab import ZoneAddrBook
from .ab_finder import AddrBookFinder



# utils/config.py
import os

# 3rd-party modules
from lxml import etree

# package modules
from ..exception import *
from .. import jxml as JXML
from .util import Util

class Config(Util):
  """
  Configuration Utilities:

    commit - commit changes
    commit_check - perform the commit check operation
    diff - return the diff string between running and candidate config
    pdiff - prints the diff string (debug/helper)
    load - load changes into the candidate config 
    lock - take an exclusive lock on the candidate config
    unlock - release the exclusive lock
    rollback - perform the load rollback command 
    
  """

  ### ---------------------------------------------------------------------------
  ### commit
  ### ---------------------------------------------------------------------------

  def commit( self, **kvargs ):
    """
    commit a configuration.  returns either :True: or
    raises an RPCError exception

    kvargs 
      confirm = [True | <timeout-minutes>]
      comment = <comment log string>
    """
    rpc_args = {}

    # if a comment is provided, then include that in the RPC

    comment = kvargs.get('comment')
    if comment: rpc_args['log'] = comment

    # if confirm is provided, then setup the RPC args
    # so that Junos will either use the default confirm
    # timeout (confirm=True) or a specific timeout
    # (confirm=<minutes>)

    confirm = kvargs.get('confirm')
    if confirm:
      rpc_args['confirmed'] = True
      confirm_val = str(confirm)
      if 'True' != confirm_val: rpc_args['confirm-timeout'] = confirm_val

    # dbl-splat the rpc_args since we want to pass key/value to metaexec
    # if there is a commit/check error, this will raise an execption

    try:
      self.rpc.commit_configuration( **rpc_args )
    except RpcError as err:        # jnpr.junos exception
      if err.rsp.find('ok') is not None:
        # this means there are warnings, but no errors
        return True
      else:
        raise CommitError(cmd=err.cmd,rsp=err.rsp)
    except Exception as err:
      # so the ncclient gives us something I don't want.  I'm going to convert
      # it and re-raise the commit error
      JXML.remove_namespaces( err.xml )
      raise CommitError( rsp=err.xml )

    return True

  ### -------------------------------------------------------------------------
  ### commit check
  ### -------------------------------------------------------------------------

  def commit_check(self):
    """
    perform a commit check.  if the commit check passes, this function
    will return :True:

    If there is a commit check error, then the RPC error reply XML
    structure will be returned
    """
    try:
      self.rpc.commit_configuration( check=True )
    except RpcError as err:        # jnpr.junos exception
      if err.rsp.find('ok') is not None:
        # this means there is a warning, but no errors
        return True
      else:
        raise CommitError(cmd=err.cmd,rsp=err.rsp)      
    except Exception as err:
      # :err: is from ncclient, so extract the XML data
      # and convert into dictionary
      return JXML.rpc_error( err.xml )

    return True

  ### -------------------------------------------------------------------------
  ### show | compare rollback <number|0*>
  ### -------------------------------------------------------------------------

  def diff( self, **kvargs ):
    """
    retrieve a diff-format report of the candidate config against
    either the current active config, or a different rollback.

    kvargs
      'rollback' is a number [0..50]
    """

    rb_id = kvargs.get('rollback', 0)
    if rb_id < 0 or rb_id > 50:
      raise ValueError("Invalid rollback #"+str(rb_id))

    rsp = self.rpc.get_configuration(dict(
      compare='rollback', rollback=str(rb_id)
      ))

    diff_txt = rsp.find('configuration-output').text
    return None if diff_txt == "\n" else diff_txt

  def pdiff(self, **kvargs):
    print self.diff(**kvargs)
    
  ### -------------------------------------------------------------------------
  ### helper on loading configs
  ### ------------------------------------------------------------------------- 

  def load(self, *vargs, **kvargs):
    """
    loads configuration into the device

    vargs (optional)
      is the content to load.  if the contents is a string, then you 
      must specify kvargs['format'].   

    kvargs['path'] 
      path to file of configuration.  the path extension will be used
      to determine the format of the contents.
        ['conf','text','txt'] is curly-text-style
        ['set'] is set-style
        ['xml'] is XML
      the format can specific set by using kvarg['format']

    kvargs['format']
      determines the format of the contents.  options are
      ['xml','set','text'] for XML/etree, set-style, curly-brace-style

    kvargs['template_path']
      path to a jinja2 template file.  used in conjection with the
      kvargs['template_vars'] option, this will perform a templating
      render and then load the result.  The template extension will
      be used to determine the format-style of the contents, or you 
      can override using kvargs['format']

    kvargs['template']
      jinja2 Template.  same description as kvargs['template_path'],
      except this option you provide the actual Template, rather than
      a path to the template file

    kvargs['template_vars']
      used in conjection with the other template options.  this option
      contains a dictionary of variables to render into the template
    """
    rpc_xattrs = {'format':'xml'}      # junos attributes, default to XML
    rpc_contents = None

    ### -----------------------------------------------------------------------
    ### private helpers ...
    ### -----------------------------------------------------------------------

    def _lformat_byext( path ):
      """ determine the format style from the file extension """
      ext = os.path.splitext(path)[1]
      if ext == '.xml': return 'xml'
      if ext in ['.conf','.text','.txt']: return 'text'
      if ext in ['.set']: return 'set'
      raise ValueError("Unknown file contents from extension: %s" % ext)

    def _lset_format(kvargs, rpc_xattrs):
      """ setup the kvargs/rpc_xattrs """
      # when format is given, setup the xml attrs appropriately
      if kvargs['format'] == 'set': 
        rpc_xattrs['action'] = 'set'
        kvargs['format'] = 'text'      
      rpc_xattrs['format'] = kvargs['format']       

    def _lset_fromfile(path):
      """ setup the kvargs/rpc_xattrs based on path """
      if 'format' not in kvargs:
        # we use the extension to determine the format
        kvargs['format'] = _lformat_byext(path)
        _lset_format( kvargs, rpc_xattrs )      

    ### -----------------------------------------------------------------------
    ### end-of: private helpers
    ### -----------------------------------------------------------------------

    if 'format' in kvargs: 
      _lset_format(kvargs, rpc_xattrs)

    ### -----------------------------------------------------------------------
    ### if contents are provided as vargs[0], then process that as XML or str
    ### -----------------------------------------------------------------------

    if len(vargs):
      # caller is providing the content directly.
      rpc_contents = vargs[0]
      if isinstance(rpc_contents,str) and not 'format' in kvargs:
        raise RuntimeError("You must define the format of the contents")
      return self.rpc.load_config( rpc_contents, **rpc_xattrs )

      #~! UNREACHABLE !~#

    ### -----------------------------------------------------------------------
    ### if path is provided, use the static-config file
    ### -----------------------------------------------------------------------

    if 'path' in kvargs:
      # then this is a static-config file.  load that as our rpc_contents
      rpc_contents = open(kvargs['path']).read()
      _lset_fromfile(kvargs['path'])
      if rpc_xattrs['format'] == 'xml':
        # covert the XML string into XML structure
        rpc_contents = etree.XML(rpc_contents)

      return self.rpc.load_config( rpc_contents, **rpc_xattrs )

      #~! UNREACHABLE !~#

    ### -----------------------------------------------------------------------
    ### if template_path is provided, then jinja2 load the template, and
    ### render the results.  if template_vars are provided, use those
    ### in the render process.
    ### -----------------------------------------------------------------------

    if 'template_path' in kvargs:
      path = kvargs['template_path']
      template = self.dev.Template(path)
      rpc_contents = template.render(kvargs.get('template_vars', {}))
      _lset_fromfile(path)
      return self.rpc.load_config( rpc_contents, **rpc_xattrs )

      #~! UNREACHABLE !~#

    ### -----------------------------------------------------------------------
    ### if template is provided, then this is a pre-loaded jinja2 Template
    ### object.  Use the template.filename to determine the format style
    ### -----------------------------------------------------------------------

    if 'template' in kvargs:
      template = kvargs['template']
      path = template.filename
      rpc_contents = template.render(kvargs.get('template_vars', {}))
      _lset_fromfile(path)
      return self.rpc.load_config( rpc_contents, **rpc_xattrs )

      #~! UNREACHABLE !~#

    raise RuntimeError("Unhandled load request")

  ### -------------------------------------------------------------------------
  ### config exclusive
  ### -------------------------------------------------------------------------

  def lock(self):
    """
    attempts an exclusive lock on the candidate configuration
    """
    try:
      self.rpc.lock_configuration()
    except Exception as err:
      # :err: is from ncclient
      raise LockError(rsp = JXML.remove_namespaces(err.xml))

    return True

  ### -------------------------------------------------------------------------
  ### releases the exclusive lock
  ### -------------------------------------------------------------------------

  def unlock( self ):
    """
    unlocks the candidate configuration
    """
    try:
      self.rpc.unlock_configuration()
    except Exception as err:
      # :err: is from ncclient
      raise UnlockError(rsp = JXML.remove_namespaces(err.xml))

    return True

  ### -------------------------------------------------------------------------
  ### rollback <number|0*>
  ### -------------------------------------------------------------------------

  def rollback( self, rb_id=0 ):
    """
    rollback the candidate config to either the last active or
    a specific rollback number.
    """

    if rb_id < 0 or rb_id > 50:
      raise ValueError("Invalid rollback #"+str(rb_id))

    self.rpc.load_configuration(dict(
      compare='rollback', rollback=str(rb_id)
    ))

    return True



# 3rd-party packages
from lxml.builder import E 

# local packages
from .. import Resource
from ... import jxml as JXML

class PolicyRule( Resource ):
  """
  [edit security policy from-zone <from_zone> to-zone <to_zone> policy <policy_name>]

  Resource name: str
    <policy_name> is the name of the policy

  Managed by: PolicyContext
    <from_zone> and <to_zone> taken from parent resource
  """

  PROPERTIES = [
    'description',
    'match_srcs',
    'match_dsts',
    'match_apps',
    'action',
    'count',
    'log_init',
    'log_close'
  ]

  ### -------------------------------------------------------------------------
  ### XML reading
  ### -------------------------------------------------------------------------

  def _xml_at_top(self):
    xml = self._parent._xml_at_top()
    xml.find('.//policy').append(E.policy(E.name( self._name )))
    return xml

  def _xml_at_res(self, xml):
    return xml.find('.//policy/policy')

  def _xml_to_py(self, as_xml, to_py ):
    Resource._r_has_xml_status( as_xml, to_py )
    Resource.copyifexists(as_xml,'description',to_py)

    x_match = as_xml.find('match')
    x_then = as_xml.find('then')

    ### collect up the 'match' criteria

    to_py['match_srcs'] = [this.text for this in x_match.findall('source-address')]
    to_py['match_dsts'] = [this.text for this in x_match.findall('destination-address')]
    to_py['match_apps'] = [this.text for this in x_match.findall('application')]

    ### collect up the 'then' criteria

    to_py['action'] = x_then.xpath('permit | reject | deny')[0].tag

    if x_then.find('count') is not None: to_py['count'] = True
    if x_then.find('log/session-init') is not None: to_py['log_init'] = True
    if x_then.find('log/session-close') is not None: to_py['log_close'] = True

  ### -------------------------------------------------------------------------
  ### XML writing
  ### -------------------------------------------------------------------------

  def _xml_change_action(self, xml):
    xml.append(E.then(E(self.should['action'])))
    return True
    
  def _xml_change_count(self, xml):
    xml.append(E.then(
      Resource.xmltag_set_or_del('count', self.should['count'])
    ))
    return True

  def _xml_change_log_init(self, xml):
    xml.append(E.then(E.log(
      Resource.xmltag_set_or_del('session-init', self.should['log_init'])
    )))
    return True

  def _xml_change_log_close(self, xml):
    xml.append(E.then(E.log(
      Resource.xmltag_set_or_del('session-close', self.should['log_close'])
    )))
    return True

  def _xml_change_match_srcs(self, xml):
    adds, dels = Resource.diff_list( self.has['match_srcs'], self.should['match_srcs'] )

    if len(adds):
      x_match = E.match()
      xml.append(x_match)
      for this in adds: x_match.append(E('source-address',E.name(this)))

    if len(dels):
      x_match = E.match()
      xml.append(x_match)
      for this in dels: x_match.append(E('source-address', JXML.DEL, E.name(this)))

    return True

  def _xml_change_match_dsts(self, xml):
    adds, dels = Resource.diff_list( self.has['match_dsts'], self.should['match_dsts'] )

    if len(adds):
      x_match = E.match()
      xml.append(x_match)
      for this in adds: x_match.append(E('destination-address',E.name(this)))

    if len(dels):
      x_match = E.match()
      xml.append(x_match)
      for this in dels: x_match.append(E('destination-address', JXML.DEL, E.name(this)))

    return True

  def _xml_change_match_apps(self, xml):
    adds, dels = Resource.diff_list( self.has['match_apps'], self.should['match_apps'] )

    if len(adds):
      x_match = E.match()
      xml.append(x_match)
      for this in adds: x_match.append(E('application',E.name(this)))

    if len(dels):
      x_match = E.match()
      xml.append(x_match)
      for this in dels: x_match.append(E('application', JXML.DEL, E.name(this)))

    return True

  ##### -----------------------------------------------------------------------
  ##### Resource List, Catalog
  ##### -- only executed by 'manager' resources
  ##### -----------------------------------------------------------------------

  def _r_list(self):
    got = self.P._xml_config_read()
    self._rlist = [ this.text for this in got.xpath('.//policy/policy/name') ]

  def _r_catalog( self ):
    got = self.D.rpc.get_config(self.P._xml_at_top())
    policies = got.find('.//security/policies/policy')
    for pol in policies.findall('policy'):
      name = pol.find('name').text
      self._rcatalog[name] = {}
      self._xml_to_py( pol, self._rcatalog[name] )
  



# 3rd-party modules
from lxml.builder import E 

# module packages
from ... import jxml as JXML
from .. import Resource

class ZoneAddrBookAddr( Resource ):
  """
  [edit security zone security-zone <zone> address-book address <name>]

  Resource name: str
    <name> is the name of the address item

  Managed by: ZoneAddrBook
    <zone> is the name of the security zone
  """

  PROPERTIES = [
    'description',
    'ip_prefix',
  ]

  def _xml_at_top(self):
    return E.security(E.zones(
      E('security-zone', 
        E.name(self.P._name),
        E('address-book',E.address(self._name))
      )
    ))

  ##### -----------------------------------------------------------------------
  ##### XML reading
  ##### -----------------------------------------------------------------------

  def _xml_at_res(self, xml):
    return xml.find('.//address-book/address')

  def _xml_to_py(self, as_xml, to_py ):
    Resource._r_has_xml_status( as_xml, to_py )
    Resource.copyifexists( as_xml, 'description', to_py )
    to_py['ip_prefix'] = as_xml.find('ip-prefix').text

  ##### -----------------------------------------------------------------------
  ##### XML writing
  ##### -----------------------------------------------------------------------

  def _xml_change_ip_prefix(self, xml):
    xml.append(E('ip-prefix', self.should['ip_prefix']))
    return True

  ##### -----------------------------------------------------------------------
  ##### Manager List, Catalog
  ##### -----------------------------------------------------------------------

  def _r_list(self):
    #  The parent keeps a property on this list, so just use it, yo!
    self._rlist = self.P['$addrs']

  def _r_catalog(self):
    get = E.security(E.zones(
      E('security-zone', 
        E.name(self.P._name),
        E('address-book', E('address'))
      )
    ))
    got = self.D.rpc.get_config( get )
    for addr in got.xpath('.//address-book/address'):
      name = addr.find('name').text
      self._rcatalog[name] = {}
      self._xml_to_py( addr, self._rcatalog[name] )      



# 3rd-party modules
from lxml.builder import E

# local module
from . import Resource
from .. import jxml as JXML

class UserSSHKey( Resource ):
  """
  [edit system login user <name> authentication <key-type> <key-value> ]

  Resource name: tuple(<key-type>, <key-value>)
    <key-type> : ['ssh-dsa', 'ssh-rsa']
    <key-value> : SSH public key string (usually something very long)

  Resource manager utilities:
    load_key - allows you to load an ssh-key from a file or str
  """

  # there are no properties, since the name <key-value> constitutes the
  # actual ssk key data, yo!

  PROPERTIES = []  

  ##### -----------------------------------------------------------------------
  ##### XML readers
  ##### -----------------------------------------------------------------------

  def _xml_at_top(self):
    key_t, key_v = self._name
    return E.system(E.login(E.user(
      E.name( self.P.name ),
      E.authentication(
        E(key_t, E.name(key_v)
      )
    ))))

  def _xml_at_res(self, xml):
    return xml.find('.//authentication/%s' % self._name[0])

  def _xml_to_py(self, has_xml, has_py ):
    Resource._r_has_xml_status( has_xml, has_py )   

  ##### -----------------------------------------------------------------------
  ##### UTILITY FUNCTIONS
  ##### -----------------------------------------------------------------------

  def load_key(self, path=None, key_value=None):
    """
    Adds a new ssh-key to the user authentication.  You can
    provide either the path to the ssh-key file, or the contents
    of they key (useful for loading the same key on many devices)

    :path: (optional)
      path to public ssh-key file on the local server,

    :key_value: (optional)
      the contents of the ssh public key
    """

    if not self.is_mgr:
      raise RuntimeError("must be a resource-manager!")

    if path is None and key_value is None:
      raise RuntimeError("You must provide either path or key_value")

    if path is not None:
      # snarf the file into key_value, yo!
      with open(path,'r') as f: key_value = f.read().strip()

    # extract some data from the key value, this will either 
    # be 'ssh-rsa' or 'ssh-dss'.  we need to decode this to set
    # the type correctly in the RPC.

    vt = key_value[0:7]  
    key_map = {'ssh-rsa':'ssh-rsa', 'ssh-dss':'ssh-dsa'}
    key_type = key_map.get(vt)
    if key_type is None:
      raise RuntimeError("Unknown ssh public key file type: %s" % vt)

    # at this point we are going to add a new key, so really what we are
    # doing is accessing a new instance of this class and 
    # doing a write, but just a touch since there are no properties, yo!

    new_key = self[(key_type, key_value)]
    return new_key.write( touch=True )

  ##### -----------------------------------------------------------------------
  ##### Manager List, Catalog
  ##### -----------------------------------------------------------------------

  def _r_list(self):
    # the key list comes from the parent object.
    self._rlist = self.P['$sshkeys']

  def _r_catalog(self):
    # no catalog but the keys
    self._rcatalog = { k:None for k in self.list }



# debuggin
from lxml import etree

# 3rd-party modules
from lxml.builder import E 

# module packages
from ... import jxml as JXML
from .. import Resource
from .zone_ab_addr import ZoneAddrBookAddr
from .zone_ab_set import ZoneAddrBookSet

class ZoneAddrBook( Resource ):
  """
  [edit security zone security-zone <zone> address-book]

  Resource name: str
    The zone name

  Resource manages:
    addr, ZoneAddrBookAddr
    set, ZoneAddrBookSet
  """

  PROPERTIES = [
    '$addrs',         # read-only addresss
    '$sets'           # read-only address-sets
  ]
  def __init__(self, junos, name=None, **kvargs ):
    if name is None:
      # resource-manager
      Resource.__init__( self, junos, name, **kvargs )
      return

    self.addr = ZoneAddrBookAddr( junos, parent=self )
    self.set = ZoneAddrBookSet( junos, parent=self )
    self._manages = ['addr','set']
    Resource.__init__( self, junos, name, **kvargs )

  def _xml_at_top(self):
    return E.security(E.zones(
      E('security-zone', 
        E.name( self._name ),
        E('address-book',
          E('address', JXML.NAMES_ONLY),
          E('address-set', JXML.NAMES_ONLY)
        )
      )
    ))

  ##### -----------------------------------------------------------------------
  ##### XML reading
  ##### -----------------------------------------------------------------------
  
  def _xml_at_res(self, xml):
    return xml.find('.//address-book')

  def _xml_to_py(self, as_xml, to_py ):
    Resource._r_has_xml_status( as_xml, to_py )
    to_py['$addrs'] = [name.text for name in as_xml.xpath('address/name')]
    to_py['$sets'] = [name.text for name in as_xml.xpath('address-set/name')]

  ##### -----------------------------------------------------------------------
  ##### Manager List, Catalog
  ##### -----------------------------------------------------------------------

  def _r_list(self):
    # this list of zone addressbooks is really just the list of zones
    got = self.R.get_zones_information(terse=True)
    zones = got.findall('zones-security/zones-security-zonename')
    self._rlist = [zone.text for zone in zones]    
    self._rlist.remove('junos-host')

  # using Resource._r_catalog()
    


# resources/srx/addrbook_finder.py
import netaddr

#from .zone import Zone
#from .addrbook import ZoneAddrBook

class AddrBookFinderResults(object):
  """
  Helper-class to hold the results of a :ZoneAddrFind.find(): invocation
  """
  def __init__(self, ab, find, results):
    self._ab = ab
    self._find = find
    self._results = results
    self.sets = []

  @property
  def lpm(self):
    """
    The longest-prefix-matching address is the last one in the results list.  This 
    fact is a result of the :ZoneAddrFinder.find(): sorted call.
    """
    return self._results[-1][0]

  @property
  def items(self):
    """
    Return a list of the matching address items and sets
    """
    return self.addrs + self.sets

  @property
  def addrs(self):
    """
    Return a list of the matching address items
    """
    # return a list of names
    return [x[0] for x in self._results]

  @property
  def matching(self):
    """
    Returns the string value of the original querried address presented to
    the find() method
    """
    return self._find

  def __repr__(self):
    """
    Provides the matching value and the zone name associated with this results
    """
    return "%s(%s in %s)" % (self.__class__.__name__, self._find, self._ab.name)

class AddrBookFinder(object):

  ### -------------------------------------------------------------------------
  ### CONSTRUCTOR
  ### -------------------------------------------------------------------------

  def __init__(self, addr_book):
    """
    addr_book
      Either a ZoneAddrBook or SharedAddrBook instance
    """
    self._ab = addr_book
    self._index = None

  def __repr__(self):
    return "AddrBookFinder(%s)" % self._ab.name

  def compile(self):
    """
    Compile a list of netaddr objects against the catalog of address items
    """
    # create a tuple of (addr-name, netaddr) for each of the items in the address-book
    self._index = [(name,netaddr.IPNetwork(addr['ip_prefix'])) for name,addr in self._ab.addr.catalog.items()]

  def find( self, addr, sets=True ):
    """
    Given an ip or ip_prefix locate the matching address book address 
    and address-set items.
    """

    # if the caller hasn't explicity invoked :compile(): to create the 
    # netaddr objects, then do that now.

    if self._index is None: self.compile()

    # convert the provided :addr: into a netaddr object and then 
    # to a subnet match to find address entries.  the matching
    # values will be sorted with longest prefix matching to be
    # last in the list

    ip = netaddr.IPNetwork(addr).ip
    in_net = lambda i: ip & i[1].netmask == i[1].network                # is ip in the subnet?
    by_pflen = lambda a,b: cmp(a[1].prefixlen, b[1].prefixlen)          # used to sort by prefix-length
    r = sorted(filter(in_net, self._index), cmp=by_pflen)               # find/sort
    if r is None: return None

    # now that we have some matching entries, we should find which
    # address-set items uses the items

    results = AddrBookFinderResults(self._ab, addr, r)
    if sets is True: results.sets = self.find_sets( results )

    # return the results object
    return results

  def find_sets(self, r):
    """
    Given a :AddrBookFinderResults: object, which contains the list of matching
    address items, locate the list of address-set objects that use those items
    """
    catalog = self._ab.set.catalog
    in_addr = lambda i: i in v['addr_list']
    sets = [k for k,v in catalog.items() if filter(in_addr, r.addrs)]
    in_set = lambda i: i in v['set_list']
    subsets = [k for k,v in catalog.items() if filter(in_set, sets)]
    return sets + subsets



# stdlib
from copy import deepcopy 

# local
from .cfgtable import CfgTable 
from .optable import OpTable
from .table import Table

from .view import View
from .viewfields import ViewFields

def FactoryCfgTable(table_name=None, data_dict={} ):
  if table_name is None: table_name = "CfgTable"
  new_cls = type(table_name, (CfgTable,), {} )
  new_cls.DEFINE = deepcopy(data_dict)
  new_cls.__module__ = __name__.replace('factory_cls','CfgTable')
  return new_cls

def FactoryOpTable(cmd, args=None, args_key=None, item=None, key=OpTable.ITEM_NAME_XPATH, view=None, table_name=None ):
  if table_name is None: table_name = "OpTable." + cmd
  new_cls = type(table_name, (OpTable,), {} )
  new_cls.GET_RPC = cmd
  new_cls.GET_ARGS = args or {}
  if args_key is not None: new_cls.GET_KEY = args_key
  new_cls.ITEM_XPATH = item
  new_cls.ITEM_NAME_XPATH = key 
  new_cls.VIEW = view
  new_cls.__module__ = __name__.replace('factory_cls','OpTable')
  return new_cls

def FactoryTable(item, key=Table.ITEM_NAME_XPATH, view=None, table_name=None):
  if table_name is None: table_name = 'Table.' + item
  new_cls = type(table_name, (Table,), {} )
  new_cls.ITEM_XPATH = item
  new_cls.ITEM_NAME_XPATH = key 
  new_cls.VIEW = view
  new_cls.__module__ = __name__.replace('factory_cls','Table')
  return new_cls

def FactoryView(fields, **kvargs):
  """
  :fields: 
    dictionary of fields, structure of which is ~internal~ and should
    not be defined explicitly. use the RunstatMaker.Fields() mechanism to 
    create theserather than hardcoding the dictionary structures; 
    since they might change over time.

  :kvargs:
    'view_name' to name the class.  this could be useful for debug
    or eventual callback mechanisms.

    'groups' is a dict of name/xpath assocaited to fields
    this technique would be used to extract fields from 
    node-set elements like port <if-device-flags>.

    'extends' names the base View class to extend.  using this
    technique you can add to existing defined Views.
  """

  view_name = kvargs.get('view_name','RunstatView')
  new_cls = type(view_name, (View,), {})

  if 'extends' in kvargs:
    base_cls = kvargs['extends']
    new_cls.FIELDS = deepcopy(base_cls.FIELDS)
    new_cls.FIELDS.update(fields)
    if 'groups' in kvargs:
      new_cls.GROUPS = deepcopy(base_cls.GROUPS)
      new_cls.GROUPS.update(kvargs['groups'])
  else:
    new_cls.FIELDS = fields
    new_cls.GROUPS = kvargs['groups'] if 'groups' in kvargs else None    

  new_cls.__module__ = __name__.replace('factory_cls','View')
  return new_cls





import re

class version_info(object):  
  def __init__(self, verstr ):
    """verstr - version string"""
    m1 = re.match('(.*?)([RBIXS])(.*)', verstr)
    self.type = m1.group(2)

    self.major = tuple(map(int,m1.group(1).split('.'))) # creates tuyple
    after_type = m1.group(3).split('.')
    self.minor = after_type[0]

    if 'X' == self.type:
      # assumes form similar to "45-D10", so extract the bits from this
      xm = re.match("(\d+)-(\w)(.*)", self.minor)
      self.minor = tuple([int(xm.group(1)), xm.group(2), int(xm.group(3))])
      if len(after_type) < 2:
        self.build = None
      else:
        self.build = int(after_type[1])
    elif 'I' == self.type:
      try:
        self.build = after_type[1]        # assumes that we have a build/spin, but not numeric
      except:
        self.build = None
    else:
      try:
        self.build = int(after_type[1])   # assumes numeric build/spin
      except:
        self.build = after_type[0]  # non-numeric

    self.as_tuple = self.major + tuple([self.minor, self.build])

  def __repr__(self):
    retstr = "junos.version_info(major={major}, type={type}, minor={minor}, build={build})".format(
      major=self.major,
      type=self.type,
      minor=self.minor,
      build=self.build
    )
    return retstr

  def _cmp_tuple(self,other):
    if self.type == 'I': raise RuntimeError("Internal Build")
    bylen = {
      2: (self.as_tuple[0:2]),
      4: self.as_tuple
    }
    return bylen[len(other)]

  def __lt__(self,other): return self._cmp_tuple(other) < other
  def __le__(self,other): return self._cmp_tuple(other) <= other
  def __gt__(self,other): return self._cmp_tuple(other) > other
  def __ge__(self,other): return self._cmp_tuple(other) >= other
  def __eq__(self,other): return self._cmp_tuple(other) == other  
  def __ne__(self,other): return self._cmp_tuple(other) != other

def _get_swver(dev,facts):
  try:
    return dev.rpc.cli("show version invoke-on all-routing-engines", format='xml')
  except:
    try:
      facts['vc_capable'] = True      
      return dev.rpc.cli("show version all-members", format='xml')
    except:
      facts['vc_capable'] = False
      return dev.rpc.get_software_information()

def software_version(junos, facts):
  
  f_persona = facts.get('personality')
  f_master = facts.get('master')
      
  x_swver = _get_swver(junos,facts)

  # ---------------------------------------------------------------------------
  # extract the version information out of the RPC response
  # ---------------------------------------------------------------------------
  
  if x_swver.tag == 'multi-routing-engine-results':
    # we need to find/identify each of the routing-engine (CPU) versions.  

    facts['2RE'] = True
    versions = []

    for re_sw in x_swver.xpath('.//software-information'):
      re_name = re_sw.xpath('preceding-sibling::re-name')[0].text

      # handle the cases where the "RE name" could be things like 
      # "FPC<n>" or "ndoe<n>", and normalize to "RE<n>".
      re_name = re.sub(r'(\w+)(\d+)','RE\\2',re_name)

      pkginfo = re_sw.xpath('package-information[name="junos"]/comment')[0].text    

      try:
        versions.append((re_name.upper(), re.findall(r'\[(.*)\]', pkginfo)[0]))
      except:
        versions.append((re_name.upper(), "0.0I0.0"))

    # now add the versions to the facts <dict>
    for re_ver in versions: facts['version_' + re_ver[0]] = re_ver[1]

    if f_master is not None:
      master = f_master[0] if isinstance(f_master,list) else f_master
      facts['version'] = facts['version_'+master]
    else:
      facts['version'] = versions[0][1]

  else:
    # single-RE
    pkginfo = x_swver.xpath('.//package-information[name = "junos"]/comment')[0].text    
    facts['version'] = re.findall(r'\[(.*)\]', pkginfo)[0]

  # ---------------------------------------------------------------------------
  # create a 'version_info' object based on the master version
  # ---------------------------------------------------------------------------

  facts['version_info'] = version_info(facts['version'])



# 3rd-party
from lxml.builder import E

# local module
from .. import Resource

class PhyPortBase( Resource ):
  """
  [edit interfaces <name>]

  Resource name: str
    <name> is the interface-name (IFD), e.g. 'ge-0/0/0'
  """

  PROPERTIES = [
    'admin',              # True
    'description',        # str
    'speed',              # ['10m','100m','1g','10g']
    'duplex',             # ['full','half']
    'mtu',                # int
    'loopback',           # True
    '$unit_count'         # number of units defined
  ]

  PORT_DUPLEX = {
    'full' : 'full-duplex',
    'half' : 'half-duplex'
  }

  @classmethod
  def _set_invert(cls,in_this,item,from_this):
    from_item = in_this[item]
    in_this[item] = [_k for _k,_v in from_this.items() if _v == from_item][0]

  ##### -----------------------------------------------------------------------
  ##### XML readers
  ##### -----------------------------------------------------------------------

  def _xml_at_top(self):
    return E.interfaces(E.interface( 
      E.name( self._name )
    ))

  def _xml_at_res(self, xml):
    return xml.find('.//interface')

  def _xml_to_py(self, has_xml, has_py ):
    # common to all subclasses
    Resource._r_has_xml_status( has_xml, has_py )
    has_py['admin'] = bool(has_xml.find('disable') is None)
    Resource.copyifexists(has_xml, 'description', has_py )
    Resource.copyifexists(has_xml, 'mtu', has_py )
    has_py['$unit_count'] = len(has_xml.findall('unit'))

  ##### -----------------------------------------------------------------------
  ##### XML writers
  ##### -----------------------------------------------------------------------

  # description handed by Resource

  def _xml_change_admin(self, xml):
    xml.append(Resource.xmltag_set_or_del('disable', (self.admin == False)))
    return True

  def _xml_change_mtu(self,xml):
    Resource.xml_set_or_delete(xml, 'mtu', self.mtu )
    return True

  ##### -----------------------------------------------------------------------
  ##### Manager List, Catalog
  ##### -----------------------------------------------------------------------

  def _r_list(self):
    got = self.R.get_interface_information(media=True, interface_name="[xgf]e*")
    self._rlist = [name.text.strip() for name in got.xpath('physical-interface/name')]    


class ViewFields(object):
  """
  Used to dynamically create a field dictionary used with the 
  RunstatView class
  """
  def __init__(self):
    self._fields = dict()

  def _prockvargs(self, field, name, **kvargs):
    if not len(kvargs): return
    field[name].update(kvargs)

  @property
  def end(self):
    return self._fields

  def str(self, name, xpath=None, **kvargs):
    """ field is a string """
    if xpath is None: xpath=name
    field = {name:{'xpath': xpath}}
    self._prockvargs( field, name, **kvargs )
    self._fields.update(field)
    return self

  def astype(self,name,xpath=None, astype=int, **kvargs):
    """
    field string value will be passed to function :astype:

    This is typically used to do simple type conversions,
    but also works really well if you set :astype: to
    a function that does a basic converstion like look
    at the value and change it to a True/False.  For
    example:

      astype=lambda x: True if x == 'enabled' else False
    """
    if xpath is None: xpath=name    
    field = {
      name: {'xpath': xpath, 'astype': astype }
    }
    self._prockvargs(field,name,**kvargs)
    self._fields.update( field )
    return self

  def int(self,name,xpath=None,**kvargs):
    """ field is an integer """
    return self.astype(name,xpath, int, **kvargs)

  def flag(self,name,xpath=None,**kvargs):
    """ 
    field is a flag, results in True/False if the xpath element exists or not 
    model this as a boolean type <bool>
    """
    return self.astype(name,xpath,bool,**kvargs)

  def table(self, name, table ):
    """ field is a RunstatTable """
    self._fields.update({
      name: {'table': table }
    })
    return self



# debuggin
from lxml import etree

# 3rd-party modules
from lxml.builder import E 

# module packages
from ... import jxml as JXML
from .. import Resource

class HostInbSvcMixin(object):

  def _xml_hook_build_change_begin(self,xml):   
    """
    HostInbSvcMixin
    """
    def _default_empty(prop_name) :
      if self.should.has_key(prop_name) and self.should[prop_name] is None:
        self.should[prop_name] = []

    _default_empty('services')
    _default_empty('protocols')

  def _xml__change__hit(self, xml, prop, ele_name):
    (adds,dels) = Resource.diff_list(self.has[prop], self.should[prop])
    if not adds and not dels: return False

    hit = E('host-inbound-traffic')
    for this in adds: hit.append(E(ele_name, E.name(this)))
    for this in dels: hit.append(E(ele_name, JXML.DEL, E.name(this)))
    xml.append(hit)
    return True

  def _xml_change_services(self, xml):
    return self._xml__change__hit( xml, 'services', 'system-services' )

  def _xml_change_protocols(self, xml):
    return self._xml__change__hit( xml, 'protocols', 'protocols' )

class ZoneInterface( HostInbSvcMixin, Resource ):
  """
  [edit security zone security-zone <zone> interfaces <name>]

  Resource name: str 
    <name> is the interface name, e.g. 'ge-0/0/0.12'

  Managed by: Zone
    <zone> is the name of the security zone    
  """

  PROPERTIES = [
    'services',
    'protocols',
  ]

  def _xml_at_top(self):
    return E.security(E.zones(
      E('security-zone', 
        E.name(self.P._name),
        E.interfaces(E.name(self._name))
      )
    ))

  ##### -----------------------------------------------------------------------
  ##### XML reading
  ##### -----------------------------------------------------------------------

  def _xml_at_res(self, xml):
    return xml.find('.//interfaces')

  def _xml_to_py(self, as_xml, to_py ):
    Resource._r_has_xml_status( as_xml, to_py )
    e = as_xml.xpath('host-inbound-traffic/system-services/name')
    to_py['services'] = [n.text for n in e]
    e = as_xml.xpath('host-inbound-traffic/protocols/name')
    to_py['protocols'] = [n.text for n in e]

  ##### -----------------------------------------------------------------------
  ##### XML writing
  ##### -----------------------------------------------------------------------

  ## handled by the mixin

  ##### -----------------------------------------------------------------------
  ##### Manager List, Catalog
  ##### -----------------------------------------------------------------------

  def _r_list(self):
    self._rlist = self.P['$ifs_list']

  # using Resource._r_catalog()


# stdlib
import warnings
from pprint import pformat
from copy import deepcopy

# 3rd-party
from lxml.builder import E 

# package modules
from .. import jxml as JXML

P_JUNOS_EXISTS = '_exists'
P_JUNOS_ACTIVE = '_active'

class Resource(object):

  PROPERTIES = [
    P_JUNOS_EXISTS,
    P_JUNOS_ACTIVE
  ]

  def __init__(self, junos, namevar=None, **kvargs ):
    """
    Resource or Resource-Manager constructor.  All managed resources
    and resource-managers inherit from this class.

    junos
      Instance of Device, this is bound to the Resource for 
      device access

    namevar
      If not None, identifies a specific resource by 'name'.  The
      format of the name is resource dependent.  Most resources take
      a single string name, while others use tuples for compound names.
      refer to each resource for the 'namevar' definition

      If namevar is None, then the instance is a Resource-Manager (RM).
      The RM is then used to select specific resources by name using
      the __getitem__ overload.

    kvargs['P'] or kvargs['parent']
      Instance to the resource parent.  This is set when resources have
      hierarchical relationships, like rules within rulesets

    kvargs['M']
      Instance to the resource manager.
    """
    self._junos = junos
    self._name = namevar
    self._parent = kvargs.get('parent') or kvargs.get('P')
    self._opts = kvargs
    self._manager = kvargs.get('M')

    if not namevar: 
      # then this is a resource-manager instance. setup the list and
      # catalog attributes, but do not load them now.  when the caller
      # invokes the properties, they will auto-load when empty.
      self._rlist = []
      self._rcatalog = {}      
      return

    # otherwise, a resource includes public attributes:

    self.properties = []
    self.properties.extend(Resource.PROPERTIES)
    if self.__class__ != Resource: 
      self.properties.extend(self.__class__.PROPERTIES)

    # if this resource manages others, then hook that 
    # into the :manages: list

    if hasattr(self,'MANAGES'):
      self._manages = self.MANAGES.keys()
      for k,v in self.MANAGES.items():
        self.__dict__[k] = v(junos,parent=self)

    # setup resource cache-attributes

    self.has = {}
    self.should = {}
    self._is_new = False

    # now load the properties from the device.
    self.read()

  ##### -----------------------------------------------------------------------
  ##### PROPERTIES
  ##### -----------------------------------------------------------------------

  @property
  def active(self):
    """
    is this resource configuration active on the Junos device?

    :RuntimeError: if invoked on a manager object
    """
    if self.is_mgr: raise RuntimeError("Not on a manager!")    
    return self.has[P_JUNOS_ACTIVE]

  @property
  def exists(self):
    """
    does this resource configuration exist on the Junos device?

    :RuntimError: if invoked on a manager
    """
    if self.is_mgr: raise RuntimeError("Not on a manager!")    
    return self.has[P_JUNOS_EXISTS]

  @property
  def is_mgr(self):
    """
    is this a resource manager?
    """    
    return (self._name == None)

  @property
  def is_new(self):
    """
    is this a new resource? that is, it does not exist
    on the Junos device when it was initally retrieved

    :RuntimeError: if invoked on a manager
    """
    if self.is_mgr: raise RuntimeError("Not on a manager!")    
    return self._is_new

  @property
  def name(self):
    """
    the name of the resource
    
    :RuntimeError: if invoked on a manager
    """
    if self.is_mgr: raise RuntimeError("Not on a manager!")    
    return self._name

  @name.setter
  def name(self, value):
    if self.is_mgr: raise RuntimeError("Not on a manager!")        
    raise AttributeError("name is currently read-only")
  
  @property
  def manages(self):
    """
    a resource may contain sub-managers for hierarchical
    oriented resources.  this method will return a list
    of manager names attached to this resource, or 
    :None: if there are not any
    """
    if hasattr(self,'_manages'): return self._manages
    return None

  @manages.setter
  def manages (self):
    raise AttributeError("read-only")

  @property
  def xml(self):
    """
    for debugging the resource XML configuration that was
    read from the Junos device
    """
    if self.is_mgr: raise RuntimeError("Not on a manager!")        
    return self._has_xml

  @property
  def list(self):
    """
    returns a list of named resources
    """
    if not self.is_mgr: raise RuntimeError("Must be a manager!")
    if not len(self._rlist): self.list_refresh()
    return self._rlist

  @property
  def catalog(self):
    """
    returns a dictionary of resources
    """
    if not self.is_mgr: raise RuntimeError("Must be a manager!")
    if not len(self._rcatalog): self.catalog_refresh()
    return self._rcatalog

  ### -------------------------------------------------------------------------
  ### shortcuts
  ### -------------------------------------------------------------------------

  @property
  def D(self):
    """ returns the Device object bound to this resource/manager """
    return self._junos

  @property 
  def R(self):
    """ returns the Device RPC meta object """
    return self._junos.rpc

  @property
  def M(self):
    """ returns the :Resource: manager associated to this resource """
    return self._manager

  @property
  def P(self):
    """ returns the parent of the associated Junos object """
    return self._parent

  ##### -----------------------------------------------------------------------
  ##### PUBLIC METHODS
  ##### -----------------------------------------------------------------------

  ### -------------------------------------------------------------------------
  ### read
  ### -------------------------------------------------------------------------

  def read(self):
    """
    read resource configuration from device
    """

    self._r_has_init()
    self._has_xml =  self._r_config_read_xml()

    if None == self._has_xml or not len(self._has_xml):
      self._is_new = True
      self._r_when_new()
      return None

    # the xml_read_parser *MUST* be implement by the 
    # resource subclass.  it is used to parse the XML
    # into native python structures.

    self._xml_to_py( self._has_xml, self.has )

    # return the python structure represntation
    return True

  ### -------------------------------------------------------------------------
  ### write
  ### -------------------------------------------------------------------------

  def write(self, **kvargs):
    """
    write resource configuration stored in :should: back to device

    kvargs['touch']
      if True, then write() will skip the check to see if any
      items exist in :should:
    """
    if self.is_mgr: raise RuntimeError("Not on a manager!")

    if not len(self.should) and 'touch' not in kvargs:
      return False

    # if this resource did not previously exist,
    # then mark it now into :should:

    if P_JUNOS_EXISTS not in self.should:
      self._r_set_exists( self.should, True )

    if self.is_new:
      self._r_set_active( self.should, True )

    # construct the XML change structure
    xml_change = self._xml_build_change()
    if None == xml_change: return False

    # write these changes to the device
    rsp = self._r_config_write_xml( xml_change )

    # copy :should: into :has: and then clear :should:
    self.has.update( self.should )
    self.should.clear()

    return True

  ### -------------------------------------------------------------------------
  ### activate
  ### -------------------------------------------------------------------------

  def activate(self):
    """
    activate resource in Junos config
    the same as the Junos config-mode "activate" command
    """
    # no action needed if it's already active 
    if self.active: return False
    self._r_set_active( self.should, True )
    return self.write()

  ### -------------------------------------------------------------------------
  ### deactivate
  ### -------------------------------------------------------------------------

  def deactivate(self):
    """
    activate resource in Junos config
    the same as the Junos config-mode "deactivate" command
    """
    # no action needed if it's already deactive
    if not self.active: return False
    self._r_set_active( self.should, False)
    return self.write()

  ### -------------------------------------------------------------------------
  ### delete
  ### -------------------------------------------------------------------------

  def delete(self):
    """
    remove configuration from Junos device
    the same as the Junos config-mode "delete" command
    """
    # cannot delete something that doesn't exist
    # @@@ should raise?

    if not self.exists: return False

    # remove the config from Junos
    xml = self._xml_edit_at_res()
    xml.attrib.update( JXML.DEL )
    self._xml_hook_on_delete( xml )
    rsp = self._r_config_write_xml( xml )

    # reset the :has: attribute
    self._r_has_init()
    return True

  ### -------------------------------------------------------------------------
  ### rename
  ### -------------------------------------------------------------------------

  def rename(self, new_name):
    """
    rename resource in Junos configuration
    the same as the Junos config-mode "rename" command
    """
    # cannot rename something that doesn't exist
    # @@@ should raise?

    if not self.exists: return False

    xml = self._xml_edit_at_res()
    xml.attrib.update( JXML.REN )
    xml.attrib.update( JXML.NAME( new_name ))

    rsp = self._r_config_write_xml( xml )
    self._name = new_name

    return True

  ### -------------------------------------------------------------------------
  ### reorder
  ### -------------------------------------------------------------------------

  def reorder( self, **kvargs ):
    """
    move the configuration within the Junos hierarcy
    the same as the Junos config-mode "insert" command

    :kvargs: 
      after="<name>"
      before="<name>"
    """
    cmd, name = next(kvargs.iteritems())
    if cmd != 'before' and cmd != 'after':
      raise ValueError("Must be either 'before' or 'after'")

    xml = self._xml_edit_at_res()
    xml.attrib.update( JXML.INSERT( cmd ))
    xml.attrib.update( JXML.NAME( name ))

    rsp = self._r_config_write_xml( xml )
    return True

  def list_refresh(self):
    """
    reloads the managed resource list from the Junos device
    """
    if not self.is_mgr: raise RuntimeError("Only on a manager!")    
    del self._rlist[:]
    self._r_list()      # invoke the specific resource method

  def catalog_refresh(self):
    """
    reloads the resource catalog from the Junos device
    """
    if not self.is_mgr: raise RuntimeError("Only on a manager!")    
    self._rcatalog.clear()
    self._r_catalog()  # invoke the specific resource method

  def _r_catalog(self):
    """
    provide a 'default' catalog creator method that simply uses
    the manager list and runs through each resource making
    a refcopy to the :has: properties
    """
    zone_list = self.list
    for name in zone_list:
      r = self[name]
      self._rcatalog[name] = r.has    


  def refresh(self):
    if not self.is_mgr: raise RuntimeError("Only on a manager!")
    self.list_refresh()
    self.catalog_refresh()
    
  def propcopy(self, p_name):
    """
    proptery from :has: to :should:

    performs a 'deepcopy' of the property; used to make
    changes to list, dict type properties
    """
    self.should[p_name] = deepcopy(self.has[p_name])
    return self.should[p_name]

  ##### -----------------------------------------------------------------------
  ##### OVERLOADS
  ##### -----------------------------------------------------------------------

    ### ---------------------------------------------------------------------
    ### ITEMS: for read/write of resource managed properties
    ### ---------------------------------------------------------------------

  def __getitem__( self, namekey ):
    """
    implements [] to obtain property value.  value will come
    from :should: if set or from :has: otherwise.
    """

    if self.is_mgr: 
      # ---------------------------------------------------------------      
      # use is resource-manager accessing specific index/name
      # to return resource instance
      # ---------------------------------------------------------------      

      self._opts['M'] = self
      if isinstance(namekey,int):
        # index, not name
        namekey = self.list[namekey]

      res = self.__class__( self._junos, namekey, **self._opts )
      return res

    # ---------------------------------------------------------------
    # use-case is resource instance for read property
    # ---------------------------------------------------------------

    if namekey in self.should:   
      return self.should[namekey]

    if namekey in self.has:    
      return self.has[namekey]

    if namekey in self.properties: 
      # it's a valid property, just not set in the resource
      return None

    raise ValueError("Unknown property request: %s" % namekey)

  def __setitem__(self, r_prop, value):
    """
    implements []= to set property value into :should:
    """
    if self.is_mgr: 
      raise RuntimeError("Not on a manager!")
    if r_prop in self.properties:
      self.should[r_prop] = value
    else:
      raise ValueError("Uknown property request: %s" % r_prop)

  def __call__(self, **kvargs):
    """
    alternative way to set property values as aggregation of
    key/value pairs.  this will automatically call :write():
    when completed.
    """
    if self.is_mgr: raise RuntimeError("Not on a manager!")
    if not kvargs: return False

    # validate property names first!
    for p_name,p_val in kvargs.items():
      if p_name not in self.properties:
        raise ValueError("Unknown property: %s" % p_name)

    # now cleared to add all the values
    self.should.update( kvargs )
    return self.write()

    ### ---------------------------------------------------------------------
    ### ATTRIBUTE: for read/write of resource managed properties
    ### ---------------------------------------------------------------------

  def __getattr__(self,namekey):
    """ 
    returns property value, accessed as attribute <resource>.<property>
    only for resource instance 
    """
    if self.is_mgr: raise RuntimeError("not on a resource-manager")
    return self[namekey]

  def __setattr__(self,name,value):
    if hasattr(self,'properties') and name in self.properties:
      # we can set this name/value in the resource property
      self[name] = value
    else:
      # pass 'up' to standard setattr method
      object.__setattr__(self,name,value)

    ### ---------------------------------------------------------------------
    ### PRINT/DEBUG
    ### ---------------------------------------------------------------------

  def __repr__(self):
    """
      stringify for debug/printing

      this will show the resource manager (class) name, 
      the resource (Junos) name, and the contents
      of the :has: dict and the contents of the :should: dict
    """
    mgr_name = self.__class__.__name__    
    return "NAME: %s: %s\nHAS: %s\nSHOULD:%s" % \
      (mgr_name, self._name, pformat(self.has), pformat(self.should)) \
      if not self.is_mgr \
      else "Resource Manager: %s" % mgr_name

    ### ---------------------------------------------------------------------
    ### ITERATOR
    ### ---------------------------------------------------------------------

  def __iter__(self):
    """ iterate through each Resource in the Manager list """
    for name in self.list:
      yield self[name]

  ##### -----------------------------------------------------------------------
  ##### XML reading 
  ##### -----------------------------------------------------------------------

  def _r_config_read_xml(self):
    """
    read the resource config from the Junos device
    """
    get = self._xml_at_top()
    self._xml_hook_read_begin( get )
    got = self._junos.rpc.get_config( get )
    return self._xml_at_res( got )

  def _xml_at_top( self ): 
    """
    ~| WARNING |~
    resource subclass *MUST* implement this!

    Create an XML structure that will be used to retrieve 
    the resource configuration from the device.
    """    
    raise RuntimeError("Resource missing method: %s" %
      self.__class__.__name__)    

  def _xml_at_res( self, xml ): 
    """
    ~| WARNING |~
    resource subclass *MUST* implement this!

    Return the XML element of the specific resource from
    the :xml: structure.  The :xml: will be the configuration
    starting at the top of the Junos config, i.e.
    <configuration>, and the resource needs to "cursor" at
    the specific resource within that structure
    """
    raise RuntimeError("Resource missing method: %s" %
      self.__class__.__name__)
  
  ##### -----------------------------------------------------------------------
  ##### XML writing
  ##### -----------------------------------------------------------------------

  def _xml_build_change(self):
    """
      iterate through the :should: properties creating the 
      necessary configuration change structure.  if there
      are no changes, then return :None:
    """
    edit_xml = self._xml_edit_at_res()

    # if there is resource hook to do something
    # before we build up the xml configuration,
    # then do that now

    self._xml_hook_build_change_begin( edit_xml )

    # if this resource should be deleted then
    # handle that case and return

    if not self.should[P_JUNOS_EXISTS]:
      self._xml_change__exists( edit_xml )
      return edit_xml

    # otherwise, this is an update, and we need to
    # construct the XML for change

    changed = False
    for r_prop in self.properties:
      if r_prop in self.should:
        edit_fn = "_xml_change_" + r_prop
        changed |= getattr(self, edit_fn)(edit_xml)

    # if there is a resource hook to do something
    # after we've run through all the property methods
    # then invoke that now

    changed |= self._xml_hook_build_change_end( edit_xml )  

    return edit_xml if changed else None

  def _r_config_write_xml(self, xml):
    """
      write the xml change to the Junos device, 
      trapping on exceptions.
    """
    top_xml = xml.getroottree().getroot()

    try:
      result = self._junos.rpc.load_config( top_xml, action='replace' )
    except Exception as err:
      # see if this is OK or just a warning
      if len(err.rsp.xpath('.//error-severity[. = "error"]')):
        raise err
      return err.rsp

    return result    

  # ---------------------------------------------------------------------------
  # XML edit cursor methods
  # ---------------------------------------------------------------------------

  def _xml_edit_at_res(self):
    return self._xml_at_res(self._xml_at_top())

  # ---------------------------------------------------------------------------
  # XML standard properties "writers"
  # ---------------------------------------------------------------------------

  def _xml_change_description(self, xml):
    Resource.xml_set_or_delete(xml, 'description', self.should['description'])
    return True

  def _xml_change__active(self, xml):
    if self.should[P_JUNOS_ACTIVE] == self.has[P_JUNOS_ACTIVE]:
      return False
    value = 'active' if self.should[P_JUNOS_ACTIVE] else 'inactive'
    xml.attrib[value] = value
    return True

  def _xml_change__exists(self, xml): 
    # if this is a change to create something new,
    # then invoke the 'on-create' hook and return 
    # the results

    if self.should[P_JUNOS_EXISTS]:
      return self._xml_hook_on_new( xml )

    # otherwise, we are deleting this resource
    xml.attrib.update( JXML.DEL )

    # now call the 'on-delete' hook and return 
    # the results

    return self._xml_hook_on_delete( xml )

  ##### -----------------------------------------------------------------------
  ##### XML HOOK methods
  ##### -----------------------------------------------------------------------

  def _xml_hook_read_begin(self, xml):
    """
    called from :_r_config_read_xml(): after call to :_xml_at_top(): and
    before the config request is made to the Junos device.  This hook allows
    the subclass to munge the XML get-request with additional items if necessary

    Returns:
      :True: when :xml: is changed
      :False: otherwise
    """
    return False

  def _xml_hook_build_change_begin(self, xml):
    """
    called from :_xml_build_change(): before the individual property 
    methods are invoked.  allows the resource to do anything, like pruning
    stub elements that were generated as part of :_xml_at_top():

    Returns:
      :True: when :xml: is changed
      :False: otherwise
    """
    return False

  def _xml_hook_build_change_end(self, xml):
    """
    called from :_xml_build_change(): after all of the properties
    methods have been processed.  

    Returns:
      :True: when :xml: is changed
      :False: otherwise
    """
    return False

  def _xml_hook_on_delete( self, xml ):
    """
    called when an XML write operation is going to delete the resource.

    Returns:
      :True: when :xml: is changed
      :False: otherwise
    """
    return False

  def _xml_hook_on_new( self, xml ): 
    """
    called when an XML write operation is going to create a new resource.

    Returns:
      :True: when :xml: is changed
      :False: otherwise
    """
    return False

  ##### -----------------------------------------------------------------------
  ##### Resource HOOK methods
  ##### -----------------------------------------------------------------------

  def _r_when_new(self): 
    """
      called by :read(): when the resource is new; i.e.
      there is no existing Junos configuration
    """
    pass

  def _r_when_delete(self):
    """
    ~| not used yet |~
    """
    pass
    
  ##### -----------------------------------------------------------------------
  ##### ~private~ resource methods
  ##### -----------------------------------------------------------------------

  def _r_set_active(self, my_props, value):
    my_props[P_JUNOS_ACTIVE] = value

  def _r_set_exists(self, my_props, value):
    my_props[P_JUNOS_EXISTS] = value

  def _r_has_init(self):
    self.has.clear()
    self.has[P_JUNOS_EXISTS] = False
    self.has[P_JUNOS_ACTIVE] = False

  @classmethod
  def _r_has_xml_status( klass, as_xml, as_py ):
    """
      set the 'exists' and 'active' :has: values
    """
    as_py[P_JUNOS_ACTIVE] = False if as_xml.attrib.get('inactive') else True
    as_py[P_JUNOS_EXISTS] = True    

  @classmethod
  def xml_set_or_delete( klass, xml, ele_name, value):
    """
    HELPER function to either set a value or remove the element
    """
    if value is not None and not isinstance(value,str): value = str(value)
    xml.append(E(ele_name,(value if value else JXML.DEL )))

  @classmethod
  def xmltag_set_or_del( klass, ele_name, value):
    """
    HELPER function creates an XML element tag read-only 
    that includes the DEL attribute depending on :value:
    """
    return E(ele_name, ({} if value else JXML.DEL ))

  @classmethod
  def copyifexists( klass, xml, ele_name, to_py, py_name=None ):
    ele_val = xml.find(ele_name)
    if ele_val != None: 
      to_py[(py_name if py_name else ele_name)] = ele_val.text.strip()

  @classmethod
  def diff_list( klass, has_list, should_list ):
    # covert lists to perform differencing
    should = set(should_list)
    has = set(has_list)

    # return lists (added, removed)
    return (list(should - has), list(has - should))  

  def _xml_list_property_add_del_names(self, xml, prop_name, element_name):
    """
    utility method use to process :list: properties.  this will add/delete
    items give the propery type and associated XML element name
    """
    (adds,dels) = Resource.diff_list( self.has.get(prop_name,[]), self.should[prop_name])
    for this in adds: xml.append(E(element_name, E.name(this)))
    for this in dels: xml.append(E(element_name, JXML.DEL, E.name(this)))    



# stdlib
import yaml
import os.path

# local module
from .factory_loader import FactoryLoader 

__all__ = ['loadyaml']

def loadyaml( path ):
  """
  Load a YAML file at :path: that contains Table and View definitions.
  Returns a <dict> of item-name anditem-class definition.  

  If you want to import these definitions directly into your namespace, (like a module)
  you would do the following:

    globals().update( loadyaml( <path-to-yaml-file> ))

  If you did not want to do this, you can access the items as the <dict>.  For 
  example, if your YAML file contained a Table called MyTable, then you could do
  something like:

    catalog = loadyaml( <path-to-yaml-file> )
    MyTable = catalog['MyTable']

    table = MyTable(dev)
    table.get()
    ...
  """
  # if no extension is given, default to '.yml'
  if os.path.splitext(path)[1] == '': path += '.yml'  
  return FactoryLoader().load( yaml.load( open(path, 'r' )))


import re
from lxml import etree
from lxml.builder import E

class _RpcMetaExec(object):

  ##### -----------------------------------------------------------------------
  ##### CONSTRUCTOR
  ##### -----------------------------------------------------------------------

  def __init__(self, junos):
    """
      ~PRIVATE CLASS~
      creates an RPC meta-executor object bound to the provided
      ez-netconf :junos: object
    """
    self._junos = junos

  ##### -----------------------------------------------------------------------
  ##### get_config
  ##### -----------------------------------------------------------------------

  def get_config( self, filter_xml=None, options={} ):
    """
      retrieve configuration from the Junos device  

      :filter_xml: is options, defines what to retrieve.  if omitted
      then the entire configuration is returned

      :options: is a dict, creates attributes for the RPC
    """
    rpc = E('get-configuration', options)

    if filter_xml != None:
      # wrap the provided filter with toplevel <configuration> if 
      # it does not already have one
      
      at_here = rpc if 'configuration' == filter_xml.tag else E('configuration')
      at_here.append( filter_xml )
      rpc.append( at_here )

    return self._junos.execute( rpc )

  ##### -----------------------------------------------------------------------
  ##### load_config
  ##### -----------------------------------------------------------------------

  def load_config( self, contents, **options ):
    """
    loads :contents: onto the Junos device, does not commit the change.

    :options: is a dictionary of XML attributes to set within the
    <load-configuration> RPC.  The :contents: are interpreted
    by the :options: as follows:

      format='text' and action='set', then :contents: is a string
      containing a series of "set" commands

      format='text', then :contents: is a string containing Junos
      configuration in curly-brace/text format

      <otherwise> :contents: is XML structure
    """
    rpc = E('load-configuration', options)

    if ('action' in options) and (options['action'] == 'set'):
      rpc.append(E('configuration-set', contents ))
    elif ('format' in options) and (options['format'] == 'text'):
      rpc.append(E('configuration-text', contents ))
    else:
      # otherwise, it's just XML Element
      etree.SubElement(rpc,'configuration').append( contents )

    return self._junos.execute( rpc )

  ##### -----------------------------------------------------------------------
  ##### cli
  ##### -----------------------------------------------------------------------

  def cli( self, command, format='text' ):
    rpc = E('command', command)
    if 'text' == format: rpc.attrib['format'] = 'text'    
    return self._junos.execute( rpc )

  ##### -----------------------------------------------------------------------
  ##### method missing
  ##### -----------------------------------------------------------------------

  def __getattr__( self, rpc_cmd_name ):
    """
      metaprograms a function to execute the :rpc_cmd_name:

      the caller will be passing (*vargs, **kvargs) on
      execution of the meta function; these are the specific
      rpc command arguments(**kvargs) and options bound
      as XML attributes (*vargs)
    """
    
    rpc_cmd = re.sub('_','-', rpc_cmd_name)

    def _exec_rpc(*vargs, **kvargs):
      # create the rpc as XML command
      rpc = etree.Element( rpc_cmd )

      # kvargs are the command parameter/values
      if kvargs:
        for arg_name, arg_value in kvargs.items():
          arg_name = re.sub('_','-',arg_name)               
          if isinstance(arg_value, (tuple, list)):
            for a in arg_value:
              arg = etree.SubElement( rpc, arg_name )
              if a != True: arg.text = a
          else:
            arg = etree.SubElement( rpc, arg_name )
            if arg_value != True: arg.text = arg_value

      # vargs[0] is a dict, command options like format='text'
      if vargs:
        for k,v in vargs[0].items():
          if v != True: rpc.attrib[k] = v

      # now invoke the command against the
      # associated :junos: device and return
      # the results per :junos:execute()

      return self._junos.execute(rpc)

    # metabind help() and the function name to the :rpc_cmd_name:
    # provided by the caller ... that's about all we can do, yo!

    _exec_rpc.__doc__ = rpc_cmd
    _exec_rpc.__name__ = rpc_cmd_name

    # return the metafunction that the caller will in-turn invoke
    return _exec_rpc

  ##### -----------------------------------------------------------------------
  ##### callable
  ##### -----------------------------------------------------------------------

  def __call__( self, rpc_cmd, **kvargs ):
    """
      callable will execute the provided :rpc_cmd: against the
      attached :junos: object and return the RPC response per 
      :junos:execute()

      kvargs is simply passed 'as-is' to :junos:execute()
    """
    return self._junos.execute( rpc_cmd, **kvargs )



# 3rd-party
from lxml.builder import E

# local
from .. import Resource
from ... import JXML
from .base import PhyPortBase

class PhyPortClassic( PhyPortBase ):

  ##### -----------------------------------------------------------------------
  ##### XML readers
  ##### -----------------------------------------------------------------------

  def _xml_to_py(self, has_xml, has_py ):
    PhyPortBase._xml_to_py(self, has_xml, has_py )

    Resource.copyifexists(has_xml, 'speed', has_py)
    Resource.copyifexists(has_xml, 'link-mode', has_py, 'duplex' )
    if has_xml.find('gigether-options/loopback') is not None: has_py['loopback'] = True
    has_py['$unit_count'] = len(has_xml.findall('unit'))

    ## normalizers
    if 'duplex' in has_py:PhyPortBase._set_invert(has_py, 'duplex', self.PORT_DUPLEX)

  ##### -----------------------------------------------------------------------
  ##### XML writers
  ##### -----------------------------------------------------------------------

  def _xml_change_speed(self, xml):
    Resource.xml_set_or_delete(xml, 'speed', self.speed)    
    return True

  def _xml_change_duplex(self, xml):
    value = self.PORT_DUPLEX.get(self.duplex)
    Resource.xml_set_or_delete(xml, 'link-mode', value)
    return True

  def _xml_change_loopback(self, xml):
    opts = E('gigether-options')
    opts.append(Resource.xmltag_set_or_del('loopback', self.loopback))
    xml.append(opts)
    return True


import re as RE

def routing_engines(junos, facts):

  re_facts = ['mastership-state','status','model','up-time','last-reboot-reason']
  re_info = junos.rpc.get_route_engine_information()

  master = []
  re_list = re_info.xpath('.//route-engine')
  if len(re_list) > 1: facts['2RE'] = True

  for re in re_list:
    x_re_name = re.xpath('ancestor::multi-routing-engine-item/re-name')

    if not x_re_name:
      # not a multi-instance routing engine platform, but could
      # have multiple RE slots
      re_name = "RE" 
      x_slot = re.find('slot')
      slot_id = x_slot.text if x_slot is not None else "0"
      re_name = re_name + slot_id
    else:
      # multi-instance routing platform
      m = RE.search('(\d)', x_re_name[0].text)
      re_name = "RE" + m.group(0)   # => RE0 | RE1

    re_fd = {}
    facts[re_name] = re_fd
    for factoid in re_facts:
      x_f = re.find(factoid)
      if x_f is not None:
        re_fd[factoid.replace('-','_')] = x_f.text

    if 'mastership_state' in re_fd:
      if facts[re_name]['mastership_state'] == 'master':
        master.append(re_name)

  # --[ end for-each 're' ]----------------------------------------------------

  len_master = len(master)
  if len_master > 1:
    facts['master'] = master
  elif len_master == 1:
    facts['master'] = master[0]



import re

def personality( junos, facts ):
  
  model = facts['model']

  examine = model if model != 'Virtual Chassis' else facts['RE0']['model']

  if re.match("^(EX)|(QFX)", examine):
    persona = 'SWITCH'
  elif examine.startswith("MX"):
    persona = 'MX'
  elif examine.startswith("vMX"):
    facts['virtual'] = True
    persona = 'MX'
  elif examine.startswith("VJX"):
    facts['virtual'] = True
    persona = 'SRX_BRANCH'
  elif examine.startswith("M"):
    persona = "M"
  elif re.match("SRX\s?(\d){4}", examine):
    persona = 'SRX_HIGHEND'    
  elif re.match("SRX\s?(\d){3}", examine):
    persona = 'SRX_BRANCH'
  elif re.search("firefly", examine, re.IGNORECASE):
    facts['virtual'] = True
    persona = 'SRX_BRANCH'
  else:
    raise RuntimeError("Unknown device persona: %s" % examine)
  
  facts['personality'] = persona



from .. import Resource
from .classic import PhyPortClassic 
from .switch import PhyPortSwitch

__all__ = ['PhyPort']

class PhyPort(object):
  def __new__(cls, dev, name=None):
    supercls = { 
      'CLASSIC' : PhyPortClassic,
      'SWITCH'  : PhyPortSwitch, 
    }.get( dev.facts['ifd_style'] )
    newcls = type(cls.__name__, (supercls,), {})
    return newcls(dev, name)





import paramiko
from select import select

_JUNOS_PROMPT = '> '
_SHELL_PROMPT = '% '
_SELECT_WAIT = 0.1
_RECVSZ = 1024

class StartShell(object):

  def __init__(self,nc):
    self._nc = nc

  def wait_for( self, this=_SHELL_PROMPT ):
    """
    wait for the result of the command, expecting :this:
    @@@ need to add a timeout safeguard
    """
    chan = self._chan
    got = []
    while True:
      rd,wr,err = select([chan],[],[],_SELECT_WAIT)
      if rd:
        data = chan.recv(_RECVSZ)
        got.append(data)
        if data.endswith(this):
          break
    return got

  def send(self,data):
    """
    send the command :data: followed by a '\n' character
    """
    self._chan.send(data)
    self._chan.send('\n')

  def open(self):
    """
    open an ssh-client connection and issue the 'start shell' command to
    drop into the Junos shell (csh).
    """
    junos = self._nc

    client = paramiko.SSHClient()
    client.load_system_host_keys()
    client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    client.connect( hostname = junos.hostname, 
      port=(22,junos._port)[junos.hostname=='localhost'],
      username=junos._auth_user,
      password=junos._auth_password,
    )

    chan = client.invoke_shell()
    self._client = client
    self._chan = chan    

    self.wait_for(_JUNOS_PROMPT)
    self.send('start shell')
    self.wait_for(_SHELL_PROMPT)

  def close(self):
    self._chan.close()
    self._client.close()

  def run(self,command,this=_SHELL_PROMPT):
    """
    run a shell command and wait for the response.  The return is a 
    tuple. The first item is True/False if exit-code is 0.  The second
    item is the output of the command.
    """
    # run the command and capture the output
    self.send( command )
    got = self.wait_for( this )

    # use $? to get the exit code of the command
    self.send('echo $?')
    rc = ''.join(self.wait_for(this))
    self.last_ok = True if rc.find('0') > 0 else False
    
    return got

  ### -------------------------------------------------------------------------
  ### CONTEXT MANAGER
  ### -------------------------------------------------------------------------

  def __enter__(self):
    self.open()
    return self

  def __exit__(self,exc_ty,exc_val,exc_tb):
    self.close()


# 3rd-party modules
from lxml.builder import E 

# module packages
from ... import Resource
from .... import jxml as JXML
from .nat_proxy_arp import NatProxyArp

class NatStaticRule( Resource ):
  """
  [edit security nat static rule-set <ruleset_name> rule <rule_name>]

  Resource namevar:
    rule_name, string.  The ruleset_name is obtained from the resource parent.

  """

  PROPERTIES = [
   "description",
   "match_dst_addr",
   "match_dst_port",
   "nat_addr",
   "nat_port",
   "proxy_interface"
  ]

  ##### -----------------------------------------------------------------------
  ##### XML read
  ##### -----------------------------------------------------------------------

  def _xml_at_top(self):
    return E.security(E.nat(E.static(
      E('rule-set',
        E.name(self.P._name),
        E.rule(E.name(self._name))
      )
    )))

  def _xml_at_res(self, xml):
    return xml.find('.//rule')

  def _xml_to_py(self, as_xml, to_py ):
    """
    converts Junos XML to native Python
    """
    Resource._r_has_xml_status( as_xml, to_py )
    Resource.copyifexists( as_xml, 'description', to_py)    
    e = as_xml.find('static-nat-rule-match')
    to_py['match_dst_addr'] = e.find('destination-address').text

  ##### -----------------------------------------------------------------------
  ##### XML write
  ##### -----------------------------------------------------------------------

  def _xml_hook_build_change_begin( self, xml ):
    if 'nat_port' not in self.should:
      # if 'nat_port' is not provided, then default to the 
      # 'match_dst_port' value
      self.should['nat_port'] = self['match_dst_port']

    if 'match_dst_addr' in self.should and 'proxy_interface' in self.has:
      # if we are changing the 'match_dst_addr' and we also have a proxy
      # interface, then we need to update the proxy_interface value to the
      # new match_dst_addr value.  start by deleting the existing one:
      namevar = (self['proxy_interface'], self.has['match_dst_addr'])
      NatProxyArp(self._junos, namevar).delete()

      if 'proxy_interface' not in self.should:
        # if the 'proxy_interface' value was not actually changed, then
        # simply copy the current one into :should:  this will trigger
        # the flush/create in the property-writer below
        self.should['proxy_interface'] = self.has['proxy_interface']

    # build up some XML that will be used by the property-writers
    match = E('static-nat-rule-match')
    xml.append(match)
    then = E.then(E('static-nat', E('prefix')))
    xml.append(then)
    self._rxml_match = match
    self._rxml_then = then.find('static-nat/prefix')

  def _xml_change_match_dst_addr(self, xml):
    self._rxml_match.append(
      E('destination-address', JXML.REPLACE, self.should['match_dst_addr'])
    )
    return True

  def _xml_change_match_dst_port(self,xml):
    self._rxml_match.append(
      E('destination-port', E.low(self.should['match_dst_port']))
    )
    return True

  def _xml_change_nat_addr(self, xml):
    self._rxml_then.append(E('addr-prefix', self.should['nat_addr']))
    return True

  def _xml_change_nat_port(self, xml):
    self._rxml_then.append(E('mapped-port', E('low', self.should['nat_port'])))
    return True

  def _xml_change_proxy_interface(self, xml):
    # this is really always going to be a 'create a new resource'.  if the
    # caller is changing the 'match_dst_addr' value, then the existing entry
    # will be removed by the "hook" function.
    namevar = (self.should['proxy_interface'], self['match_dst_addr'])
    parp = NatProxyArp(self._junos, namevar)
    parp.write(touch=True)
    return True

  ##### -----------------------------------------------------------------------
  ##### Manager List, Catalog
  ##### -----------------------------------------------------------------------

  def _r_list(self):
    self._rlist = self.P['rules']

  def _r_catalog(self):
    get = E.security(E.nat(E.static(
      E('rule-set',
        E.name(self.P._name),
      )
    )))
    got = self.D.rpc.get_config( get )
    for rule in got.xpath('.//rule'):
      name = rule.find('name').text
      self._rcatalog[name] = {}
      self._xml_to_py( rule, self._rcatalog[name] )
